# Auto generated from test.yaml by pythongen.py version: 0.9.0
# Generation date: 2023-08-17T14:01:39
# Schema: microscopemetrics_samples_field_illumination_schema
#
# id: https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml
# description:
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import re
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List, Optional, Union

from jsonasobj2 import JsonObj, as_dict
from linkml_runtime.linkml_model.meta import (
    EnumDefinition,
    PermissibleValue,
    PvFormulaOptions,
)
from linkml_runtime.linkml_model.types import Float
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.dataclass_extensions_376 import (
    dataclasses_init_fn_with_kwargs,
)
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from linkml_runtime.utils.formatutils import camelcase, sfx, underscore
from linkml_runtime.utils.metamodelcore import bnode, empty_dict, empty_list
from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.yamlutils import (
    YAMLRoot,
    extended_float,
    extended_int,
    extended_str,
)
from rdflib import Namespace, URIRef

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LINKML = CurieNamespace("linkml", "https://w3id.org/linkml/")
DEFAULT_ = CurieNamespace(
    "",
    "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/",
)


# Types

# Class references


@dataclass
class MyClass(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/MyClass"
    )
    class_class_curie: ClassVar[str] = None
    class_name: ClassVar[str] = "MyClass"
    class_model_uri: ClassVar[URIRef] = URIRef(
        "https://github.com/MontpellierRessourcesImagerie/microscope-metrics/blob/main/src/microscopemetrics/data_schema/samples/field_illumination_schema.yaml/MyClass"
    )

    my_slot: Optional[Union[float, List[float]]] = 0.3

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.my_slot, list):
            self.my_slot = [self.my_slot] if self.my_slot is not None else []
        self.my_slot = [v if isinstance(v, float) else float(v) for v in self.my_slot]

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass


slots.my_slot = Slot(
    uri=DEFAULT_.my_slot,
    name="my_slot",
    curie=DEFAULT_.curie("my_slot"),
    model_uri=DEFAULT_.my_slot,
    domain=None,
    range=Optional[Union[float, List[float]]],
)



# import random
#
# import numpy as np
# import pandas as pd
#
# try:
#     from hypothesis import assume
#     from hypothesis import strategies as st
# except ImportError as e:
#     raise ImportError(
#         "In order to run the strategies you need to install the test extras. Run `pip install microscopemetrics[test]`."
#     ) from e
# import microscopemetrics_schema.strategies.analyses as st_mm_analyses_schema
# from skimage.exposure import rescale_intensity as skimage_rescale_intensity
# from skimage.filters import gaussian as skimage_gaussian
# from skimage.util import random_noise as skimage_random_noise
#
# from microscopemetrics.analyses import numpy_to_mm_image
#
#
# # Strategies for user experiment
# def _gen_user_experiment_channel(
#     z_shape: int,
#     y_shape: int,
#     x_shape: int,
#     t_shape: int,
#     lines_nr: int,
#     sigma: float,
#     target_min_intensity: float,
#     target_max_intensity: float,
#     do_noise: bool,
#     signal: int,
# ) -> np.ndarray:
#     # Generate the channel as float64
#     channel = np.zeros(shape=(t_shape, z_shape, y_shape, x_shape), dtype="float64")
#
#     # draw lines_nr lines through the center of the image in z across the y and x dimensions
#     for i in range(lines_nr):
#         channel[
#             :,  # Time
#             int(channel.shape[0] * 0.5),  # Z
#             :,  # Y
#             int(channel.shape[2] * i / lines_nr),  # X
#         ] = 1.0
#         channel[
#             :,
#             int(channel.shape[0] * 0.5),
#             int(channel.shape[1] * i / lines_nr),
#             :,
#         ] = 1.0
#         channel[
#             :,
#             :,
#             int(channel.shape[1] * i / lines_nr),
#             int(channel.shape[1] * i / lines_nr),
#         ] = 1.0
#
#     channel = skimage_gaussian(
#         channel,
#         sigma=sigma,
#         mode="constant",
#         preserve_range=True,
#     )
#
#     # Normalize channel intensity to be between target_min_intensity and target_max_intensity
#     # Saturation point is at 1.0 when we rescale later to the target dtype
#     channel = skimage_rescale_intensity(
#         channel, out_range=(target_min_intensity, target_max_intensity)
#     )
#
#     if do_noise:
#         # The noise on a 1.0 intensity image is too strong, so we rescale the image to
#         # the defined signal and then rescale it back to the target intensity
#         channel = channel * signal
#         channel = skimage_random_noise(channel, mode="poisson", clip=False)
#         channel = channel / signal
#
#     return channel
#
#
# def _gen_user_experiment_image(
#     z_shape: int,
#     y_shape: int,
#     x_shape: int,
#     c_shape: int,
#     lines_nr: list[int],
#     sigma: list[float],
#     target_min_intensity: list[float],
#     target_max_intensity: list[float],
#     do_noise: bool,
#     signal: list[int],
#     dtype: np.dtype,
#     t_shape: int = 1,  # Added the time dimension here. Not used in the function yet
# ) -> np.ndarray:
#     # Generate the image as float64
#     image = np.zeros(shape=(t_shape, z_shape, y_shape, x_shape, c_shape), dtype="float64")
#
#     for ch in range(c_shape):
#         image[..., ch] = _gen_user_experiment_channel(
#             z_shape=z_shape,
#             y_shape=y_shape,
#             x_shape=x_shape,
#             t_shape=t_shape,
#             lines_nr=lines_nr[ch],
#             sigma=sigma[ch],
#             target_min_intensity=target_min_intensity[ch],
#             target_max_intensity=target_max_intensity[ch],
#             do_noise=do_noise,
#             signal=signal[ch],
#         )
#
#     # Rescale to the target dtype
#     image = np.clip(image, None, 1)
#     image = skimage_rescale_intensity(image, in_range=(0.0, 1.0), out_range=dtype)
#
#     return image
#
#
# @st.composite
# def st_user_experiment_test_data(
#     draw,
#     nr_images=st.integers(min_value=1, max_value=3),
#     z_image_shape=st.integers(min_value=2, max_value=50),
#     y_image_shape=st.integers(min_value=512, max_value=1024),
#     x_image_shape=st.integers(min_value=512, max_value=1024),
#     c_image_shape=st.integers(min_value=1, max_value=3),
#     lines_nr=st.integers(min_value=3, max_value=7),
#     dtype=st.sampled_from([np.uint8, np.uint16, np.float32]),
#     do_noise=st.just(True),
#     signal=st.floats(min_value=20.0, max_value=1000.0),
#     target_min_intensity=st.floats(min_value=0.001, max_value=0.1),
#     target_max_intensity=st.floats(min_value=0.5, max_value=0.9),
#     sigma=st.floats(min_value=2.0, max_value=10.0),
#     nr_orthogonal_rois=st.integers(min_value=0, max_value=3),
#     nr_profile_rois=st.integers(min_value=0, max_value=3),
# ) -> dict:
#     output = {
#         "images": [],
#         "lines_nr": [],
#         "orthogonal_rois": [],
#         "profile_rois": [],
#         "do_noise": [],
#         "signals": [],
#     }
#
#     c_image_shape = draw(c_image_shape)
#     z_image_shape = draw(z_image_shape)
#     y_image_shape = draw(y_image_shape)
#     x_image_shape = draw(x_image_shape)
#
#     do_noise = draw(do_noise)
#
#     dtype = draw(dtype)
#
#     for _ in range(draw(nr_images)):
#         image_target_min_intensities = [draw(target_min_intensity) for _ in range(c_image_shape)]
#         image_target_max_intensities = [draw(target_max_intensity) for _ in range(c_image_shape)]
#         image_sigma = [draw(sigma) for _ in range(c_image_shape)]
#         image_signals = [draw(signal) for _ in range(c_image_shape)]
#         image_lines_nr = [draw(lines_nr) for _ in range(c_image_shape)]
#
#         image = _gen_user_experiment_image(
#             z_shape=z_image_shape,
#             y_shape=y_image_shape,
#             x_shape=x_image_shape,
#             c_shape=c_image_shape,
#             lines_nr=image_lines_nr,
#             sigma=image_sigma,
#             target_min_intensity=image_target_min_intensities,
#             target_max_intensity=image_target_max_intensities,
#             do_noise=do_noise,
#             signal=image_signals,
#             dtype=dtype,
#         )
#
#         orthogonal_rois = []
#         for _ in range(draw(nr_orthogonal_rois)):
#             ortho_roi = st_mm_schema.st_mm_roi(
#                 shapes=[
#                     st_mm_schema.st_mm_point(
#                         c=st.just(None),
#                         t=st.just(None),
#                         z=st.floats(
#                             min_value=0.0,
#                             max_value=z_image_shape - 1,
#                             exclude_min=True,
#                             exclude_max=True,
#                         ),
#                         y=st.floats(
#                             min_value=0.0,
#                             max_value=y_image_shape - 1,
#                             exclude_min=True,
#                             exclude_max=True,
#                         ),
#                         x=st.floats(
#                             min_value=0.0,
#                             max_value=x_image_shape - 1,
#                             exclude_min=True,
#                             exclude_max=True,
#                         ),
#                     )
#                 ]
#             )
#             orthogonal_rois.append(ortho_roi)
#
#         profile_rois = []
#         for _ in range(draw(nr_profile_rois)):
#             profile_roi = st_mm_schema.st_mm_roi(
#                 shapes=[
#                     st_mm_schema.st_mm_line(
#                         c=st.just(None),
#                         t=st.just(None),
#                         z=st.floats(
#                             min_value=0.0,
#                             max_value=z_image_shape - 1,
#                             exclude_min=True,
#                             exclude_max=True,
#                         ),
#                         x1=st.floats(
#                             min_value=0.0,
#                             max_value=x_image_shape / 2,
#                             exclude_min=True,
#                             exclude_max=True,
#                         ),
#                         x2=st.floats(
#                             min_value=x_image_shape / 2,
#                             max_value=x_image_shape - 1,
#                             exclude_min=True,
#                             exclude_max=True,
#                         ),
#                         y1=st.floats(
#                             min_value=0.0,
#                             max_value=y_image_shape / 2,
#                             exclude_min=True,
#                             exclude_max=True,
#                         ),
#                         y2=st.floats(
#                             min_value=y_image_shape / 2,
#                             max_value=y_image_shape - 1,
#                             exclude_min=True,
#                             exclude_max=True,
#                         ),
#                     )
#                 ]
#             )
#
#         output["images"].append(image)
#         output["lines_nr"].append(image_lines_nr)
#         output["do_noise"].append(do_noise)
#         output["signals"].append(image_signals)
#         output["orthogonal_rois"].append(orthogonal_rois)
#         output["profile_rois"].append(profile_rois)
#
#     return output
#
#
# @st.composite
# def st_user_experiment_dataset(
#     draw,
#     unprocessed_dataset=st_mm_analyses_schema.st_mm_user_experiment_unprocessed_dataset(),
#     test_data=st_user_experiment_test_data(),
# ):
#     test_data = draw(test_data)
#     user_experiment_unprocessed_dataset = draw(unprocessed_dataset)
#
#     # the unprocessed dataset contains one empty image, we need to delete it
#     user_experiment_unprocessed_dataset.input_data.user_experiment_images = []
#
#     for i, image in enumerate(test_data.pop("images")):
#         mm_image = numpy_to_mm_image(
#             array=image,
#             name=f"UE_image_{i}",
#             channel_names=[f"Channel_{i}{j}" for j in range(image.shape[-1])],
#         )
#
#         orthogonal_rois = test_data["orthogonal_rois"][i]
#         for ortho_roi in orthogonal_rois:
#             ortho_roi.linked_references.append(mm_image.data_reference)
#
#         profile_rois = test_data["profile_rois"][i]
#         for prof_roi in profile_rois:
#             prof_roi.linked_references.append(mm_image.data_reference)
#
#         user_experiment_unprocessed_dataset.input_data.user_experiment_images.append(mm_image)
#         user_experiment_unprocessed_dataset.input_data.orthogonal_rois.extend(orthogonal_rois)
#         user_experiment_unprocessed_dataset.input_data.profile_rois.extend(profile_rois)
#
#     # Setting the bit depth to the data type of the image
#     image_dtype = {
#         a.array_data.dtype
#         for a in user_experiment_unprocessed_dataset.input_data.user_experiment_images
#     }
#     if len(image_dtype) != 1:
#         raise ValueError("All images should have the same data type")
#     image_dtype = image_dtype.pop()
#     if np.issubdtype(image_dtype, np.integer):
#         user_experiment_unprocessed_dataset.input_parameters.bit_depth = np.iinfo(image_dtype).bits
#     elif np.issubdtype(image_dtype, np.floating):
#         user_experiment_unprocessed_dataset.input_parameters.bit_depth = np.finfo(image_dtype).bits
#     else:
#         user_experiment_unprocessed_dataset.input_parameters.bit_depth = None
#
#     return {
#         "unprocessed_dataset": user_experiment_unprocessed_dataset,
#         "expected_output": test_data,
#     }



from datetime import datetime

import microscopemetrics_schema.datamodel as mm_schema
import numpy as np
import pandas as pd
from scipy import ndimage, signal
from skimage.exposure import rescale_intensity
from skimage.feature import peak_local_max
from skimage.filters import gaussian

import microscopemetrics as mm
from microscopemetrics.analyses import tools as mm_tools

MAX_NR_PEAKS = 100


def _add_column_name_level(df: pd.DataFrame, level_name: str, level_value: str):
    if isinstance(df.columns, pd.MultiIndex):
        new_columns = pd.MultiIndex.from_tuples(
            [(level_value, *col) for col in df.columns],
            names=[level_name] + list(df.columns.names),
        )
    else:
        new_columns = pd.MultiIndex.from_tuples(
            [(level_value, col) for col in df.columns],
            names=[level_name, df.columns.name],
        )
    df.columns = new_columns


def _add_row_index_level(df: pd.DataFrame, level_name: str, level_value: str):
    if isinstance(df.index, pd.MultiIndex):
        new_index = pd.MultiIndex.from_tuples(
            [(level_value, *row) for row in df.index],
            names=[level_name] + list(df.index.names),
        )
    else:
        new_index = pd.MultiIndex.from_tuples(
            [(level_value, row) for row in df.index], names=[level_name, df.index.name]
        )
    df.index = new_index


def _concatenate_index_levels(index_names, index_values, pattern="{level_name}-{level_value}_"):
    concatenated_str = "".join(
        pattern.format(level_name=level_name, level_value=level_value)
        for level_name, level_value in zip(index_names, index_values)
    )

    return concatenated_str.rstrip("_")


def _average_beads(group: pd.DataFrame) -> pd.Series:
    """
    Averages the beads in the list by first aligning them to the center of the image and then averaging them.
    """
    dtype = {bead.dtype for bead in group.beads}
    if len(dtype) > 1:
        raise ValueError("All beads must have the same dtype.")
    aligned_beads = [
        ndimage.shift(row.beads, _find_bead_shifts(row.beads), mode="nearest", order=1)
        for row in group.itertuples()
        if row.considered_valid
    ]
    if not aligned_beads:
        mm.logger.warning("No valid beads to average.")
        return pd.Series({"average_bead": np.nan})
    if len(aligned_beads) < 2:
        mm.logger.warning("Less than 2 beads to average.")
    mm.logger.info(f"Averaging {len(aligned_beads)} beads")

    return pd.Series(
        {
            "average_bead": np.mean(aligned_beads, axis=0).astype(dtype.pop()),
        }
    )


def _find_bead_shifts(data1, data2=None):
    """Cross-correlates two 2D or 3D arrays and returns the shifts.
    If a second array is not provided, a Gaussian is used as a reference."""
    if data2 is None:
        data2 = np.zeros_like(data1)
        slices = []
        for dim in data2.shape:
            if dim % 2 == 0:
                slices.append(slice(dim // 2))
            else:
                slices.append(slice(dim // 2, dim // 2 + 1))
        data2[tuple(slices)] = 1
        data2 = gaussian(data2, sigma=1, preserve_range=True)

    if data1.ndim != data2.ndim:
        raise ValueError("Data1 and Data2 must have the same number of dimensions.")

    corr_arr = signal.correlate(data1, data2, mode="same")

    profiles = []
    max_pos = np.unravel_index(np.argmax(corr_arr), corr_arr.shape)
    for dim in range(corr_arr.ndim):
        pos_slices = [slice(pos, pos + 1) for pos in max_pos]
        pos_slices[dim] = slice(None)
        profiles.append(np.squeeze(corr_arr[tuple(pos_slices)]))

    return tuple(
        mm_tools.fit_gaussian(profile)[3][2] - profile.shape[0] // 2 for profile in profiles
    )


def _estimate_threshold(image, tolerance=2):
    """
    This function is finding a range of thresholds for which there is not much
    difference in the number of peaks found.
    """
    threshold_range = np.linspace(0.0, 1.0, 11)
    peak_counts = []

    for threshold in threshold_range:
        peak_counts.append(len(peak_local_max(image=image, threshold_abs=threshold)))

    peak_counts_diff = np.abs(np.diff(peak_counts))
    peak_counts_mask = peak_counts_diff <= tolerance

    plateau_length = 0
    plateau_start = None
    current_length = 0
    current_start = None

    for i, is_plateau in enumerate(peak_counts_mask):
        if is_plateau:
            if current_length == 0:
                current_start = i
            current_length += 1
        else:
            if current_length >= plateau_length:
                plateau_length = current_length
                plateau_start = current_start
            current_length = 0

    # Final check inc ase the plateau extends to the end
    if current_length > plateau_length:
        plateau_length = current_length
        plateau_start = current_start

    if plateau_start is None:
        return None  # No threshold found

    # We return the threshold that best matches the higher end of the plateau
    return threshold_range[plateau_start + int(plateau_length * 0.8)]


def _calculate_bead_intensity_outliers(
    bead_properties: pd.DataFrame, robust_z_score_threshold: float
) -> None:
    bead_properties["max_intensity_robust_z_score"] = pd.Series(dtype="float")
    bead_properties["std_intensity_robust_z_score"] = pd.Series(dtype="float")
    bead_properties["considered_intensity_outlier"] = pd.Series(dtype="bool")

    if len(bead_properties[bead_properties.considered_valid]) == 1:
        bead_properties["max_intensity_robust_z_score"] = 0
        bead_properties["std_intensity_robust_z_score"] = 0
        bead_properties["considered_intensity_outlier"] = False
    else:
        max_int_mean = bead_properties[bead_properties.considered_valid]["intensity_max"].mean()
        max_int_median = bead_properties[bead_properties.considered_valid]["intensity_max"].median()
        max_int_mad = (
            (bead_properties[bead_properties.considered_valid]["intensity_max"] - max_int_mean)
            .abs()
            .mean()
        )

        std_int_mean = bead_properties[bead_properties.considered_valid]["intensity_std"].mean()
        std_int_median = bead_properties[bead_properties.considered_valid]["intensity_std"].median()
        std_int_mad = (
            (bead_properties[bead_properties.considered_valid]["intensity_std"] - std_int_mean)
            .abs()
            .mean()
        )

        bead_properties["max_intensity_robust_z_score"] = (
            0.6745 * (bead_properties["intensity_max"] - max_int_median) / max_int_mad
        )
        bead_properties["std_intensity_robust_z_score"] = (
            0.6745 * (bead_properties["intensity_std"] - std_int_median) / std_int_mad
        )

        if 1 < len(bead_properties[bead_properties.considered_valid]) < 6:
            bead_properties["considered_intensity_outlier"] = False
        else:
            bead_properties["considered_intensity_outlier"] = (
                # abs(bead_positions["max_intensity_robust_z_score"]) > robust_z_score_threshold
                abs(bead_properties["std_intensity_robust_z_score"])
                > robust_z_score_threshold
            )

    bead_properties["considered_intensity_outlier"] = bead_properties[
        "considered_intensity_outlier"
    ].astype(bool)


def _generate_key_measurements(bead_properties, average_bead_properties):
    measurement_aggregation_columns = [
        "channel_name",
        "channel_nr",
        "intensity_max",
        "intensity_min",
        "intensity_std",
        "fit_r2_z",
        "fit_r2_y",
        "fit_r2_x",
        "fwhm_pixel_z",
        "fwhm_pixel_y",
        "fwhm_pixel_x",
        "fwhm_lateral_asymmetry_ratio",
        "fwhm_micron_z",
        "fwhm_micron_y",
        "fwhm_micron_x",
    ]
    count_aggregation_columns = [
        "channel_name",
        "channel_nr",
        "total_bead",
        "considered_valid",
        "considered_self_proximity",
        "considered_lateral_edge",
        "considered_axial_edge",
        "considered_intensity_outlier",
        "considered_bad_fit_z",
        "considered_bad_fit_y",
        "considered_bad_fit_x",
    ]

    reindex_bead_properties_df = bead_properties.reset_index()

    # Add a all True column to count the total number of beads
    reindex_bead_properties_df["total_bead"] = True

    # We aggregate counts for each channel on beads according to their status
    channel_counts = (
        reindex_bead_properties_df[count_aggregation_columns]
        .groupby(["channel_nr", "channel_name"])
        .agg(["sum"])
    )
    channel_counts.columns = [
        "_".join((col[0], "count")).strip() for col in channel_counts.columns.values
    ]

    # We aggregate measurements for each channel only on beads considered valid
    valid_bead_properties_df = reindex_bead_properties_df[
        reindex_bead_properties_df.considered_valid
    ]
    channel_measurements = (
        valid_bead_properties_df[measurement_aggregation_columns]
        .groupby(["channel_nr", "channel_name"])
        .agg(["mean", "median", "std"])
    )
    channel_measurements.columns = [
        "_".join(col).strip() for col in channel_measurements.columns.values
    ]

    average_bead_properties = average_bead_properties.add_prefix("average_bead_")

    key_measurements = pd.concat(
        [channel_counts, channel_measurements, average_bead_properties], axis=1
    )
    key_measurements.reset_index(inplace=True)

    return key_measurements


def _process_bead(bead: np.ndarray, voxel_size_micron: tuple[float, float, float]):
    if not isinstance(bead, np.ndarray) and np.isnan(bead):
        return {
            "z_raw": np.nan,
            "z_fitted": np.nan,
            "y_raw": np.nan,
            "y_fitted": np.nan,
            "x_raw": np.nan,
            "x_fitted": np.nan,
            "fit_r2_z": np.nan,
            "fit_r2_y": np.nan,
            "fit_r2_x": np.nan,
            "fwhm_pixel_z": np.nan,
            "fwhm_pixel_y": np.nan,
            "fwhm_pixel_x": np.nan,
            "fwhm_micron_z": np.nan,
            "fwhm_micron_y": np.nan,
            "fwhm_micron_x": np.nan,
            "fwhm_lateral_asymmetry_ratio": np.nan,
            "considered_axial_edge": np.nan,
            "intensity_integrated": np.nan,
            "intensity_max": np.nan,
            "intensity_min": np.nan,
            "intensity_std": np.nan,
        }
    # TODO: This can be shortened
    # Find the strongest sections to generate profiles
    z_max = np.max(bead, axis=(1, 2))
    z_focus = np.argmax(z_max)
    y_max = np.max(bead, axis=(0, 2))
    y_focus = np.argmax(y_max)
    x_max = np.max(bead, axis=(0, 1))
    x_focus = np.argmax(x_max)

    # Generate profiles
    profile_z_raw = np.squeeze(bead[:, y_focus, x_focus])
    profile_y_raw = np.squeeze(bead[z_focus, :, x_focus])
    profile_x_raw = np.squeeze(bead[z_focus, y_focus, :])

    # Normalize the profiles and subtract the background
    profile_z_raw = (profile_z_raw - profile_z_raw.min()) / (
        profile_z_raw.max() - profile_z_raw.min()
    )
    profile_y_raw = (profile_y_raw - profile_y_raw.min()) / (
        profile_y_raw.max() - profile_y_raw.min()
    )
    profile_x_raw = (profile_x_raw - profile_x_raw.min()) / (
        profile_x_raw.max() - profile_x_raw.min()
    )

    # Fitting the profiles
    profile_z_fitted, r2_z, fwhm_z, (center_pos_z, _) = mm_tools.fit_airy(profile_z_raw)
    profile_y_fitted, r2_y, fwhm_y, (center_pos_y, _) = mm_tools.fit_airy(profile_y_raw)
    profile_x_fitted, r2_x, fwhm_x, (center_pos_x, _) = mm_tools.fit_airy(profile_x_raw)

    fwhm_lateral_asymmetry_ratio = max(fwhm_y, fwhm_x) / min(fwhm_y, fwhm_x)

    if all(voxel_size_micron):
        fwhm_micron_z = fwhm_z * voxel_size_micron[0]
        fwhm_micron_y = fwhm_y * voxel_size_micron[1]
        fwhm_micron_x = fwhm_x * voxel_size_micron[2]
    else:
        fwhm_micron_z = np.nan
        fwhm_micron_y = np.nan
        fwhm_micron_x = np.nan

    considered_axial_edge = (
        # TODO: review the tolerance values for considering beads at axial edge
        # a 2.5x FWHM tolerance is used and that is arbitrary and rather low
        center_pos_z < fwhm_z * 2.5
        or profile_z_raw.shape[0] - center_pos_z < fwhm_z * 2.5
    )

    intensity_max = bead.max()
    intensity_min = bead.min()
    intensity_std = bead.std()
    intensity_integrated = (bead - intensity_min).sum()

    return {
        "z_raw": profile_z_raw,
        "z_fitted": profile_z_fitted,
        "y_raw": profile_y_raw,
        "y_fitted": profile_y_fitted,
        "x_raw": profile_x_raw,
        "x_fitted": profile_x_fitted,
        "fit_r2_z": r2_z,
        "fit_r2_y": r2_y,
        "fit_r2_x": r2_x,
        "fwhm_pixel_z": fwhm_z,
        "fwhm_pixel_y": fwhm_y,
        "fwhm_pixel_x": fwhm_x,
        "fwhm_micron_z": fwhm_micron_z,
        "fwhm_micron_y": fwhm_micron_y,
        "fwhm_micron_x": fwhm_micron_x,
        "fwhm_lateral_asymmetry_ratio": fwhm_lateral_asymmetry_ratio,
        "considered_axial_edge": considered_axial_edge,
        "intensity_integrated": intensity_integrated,
        "intensity_max": intensity_max,
        "intensity_min": intensity_min,
        "intensity_std": intensity_std,
    }


def _find_beads(
    channel: np.ndarray,
    sigma: tuple[float, float, float],
    min_distance: float,
    do_noisy: bool = True,
):
    """
    This function finds the beads in a channel by applying a Gaussian filter and then finding the local maxima.
    """
    mm.logger.debug("Finding beads in channel...")

    if all(sigma):
        mm.logger.debug(f"Gaussian filter applied with sigma {sigma}")
        channel_gauss = gaussian(image=channel, sigma=sigma)
    else:
        mm.logger.debug("No Gaussian filter applied")
        channel_gauss = channel

    # We find the beads in the MIP for performance and to avoid anisotropy issues in the axial direction
    channel_gauss_mip = np.max(channel_gauss, axis=0)

    # We rescale the image intensities to work with relative thresholds
    channel_gauss_mip = rescale_intensity(channel_gauss_mip, out_range=np.float32)

    # Estimate a relative threshold for the peak_local_max function
    if do_noisy:
        threshold = _estimate_threshold(channel_gauss_mip)
        if threshold is None:
            mm.logger.error("No optimal noisy beads threshold found for the channel.")
            threshold = 0.6
    else:
        threshold = 0.3

    # We assume that reaching a maximum number of beads is a sign of a bad thresholding
    # and noise in the image. We report this as an error.
    max_num_peaks = MAX_NR_PEAKS

    # Find bead centers
    positions_all = peak_local_max(
        image=channel_gauss_mip,
        threshold_abs=threshold,
        num_peaks=max_num_peaks,
    )
    if len(positions_all) == max_num_peaks:
        mm.logger.warning(f"Reached the maximum number of peaks ({max_num_peaks}) in the image. ")
        if not do_noisy:  # We did not do noisy this time
            mm.logger.info("Trying noisy beads detected.")
            return _find_beads(channel, sigma, min_distance, do_noisy=True)
        else:
            mm.logger.error("Too many beads detected. The image is probably too noisy.")
            raise mm.AnalysisError("Too many beads detected. The image is probably too noisy.")

    positions_all_not_proximity_not_edge = peak_local_max(
        image=channel_gauss_mip,
        threshold_abs=threshold,
        min_distance=int(min_distance),
        exclude_border=(int(1 + min_distance // 2), int(1 + min_distance // 2)),
        num_peaks=max_num_peaks,
        p_norm=2,
    )
    positions_all_not_proximity = peak_local_max(
        image=channel_gauss_mip,
        threshold_abs=threshold,
        min_distance=int(min_distance),
        exclude_border=False,
        num_peaks=max_num_peaks,
        p_norm=2,
    )

    # We convert the arrays into sets to perform easier to follow filtering logic
    positions_all = set(map(tuple, positions_all))
    positions_all_not_proximity_not_edge = set(map(tuple, positions_all_not_proximity_not_edge))
    positions_all_not_proximity = set(map(tuple, positions_all_not_proximity))

    # positions_proximity_edge = positions_all - positions_all_not_proximity_not_edge
    positions_edge = positions_all_not_proximity - positions_all_not_proximity_not_edge
    positions_proximity = positions_all - positions_all_not_proximity

    positions_df = pd.DataFrame(
        positions_all,
        columns=["center_y", "center_x"],
        index=pd.Index(range(len(positions_all)), name="bead_id"),
    )
    positions_df["center_z"] = positions_df.apply(
        lambda row: int(channel_gauss[:, row["center_y"], row["center_x"]].argmax()),
        axis=1,
    )

    positions_df["considered_valid"] = positions_df.apply(
        lambda row: (row["center_y"], row["center_x"]) in positions_all_not_proximity_not_edge,
        axis=1,
    )
    positions_df["considered_self_proximity"] = positions_df.apply(
        lambda row: (row["center_y"], row["center_x"]) in positions_proximity, axis=1
    )
    positions_df["considered_lateral_edge"] = positions_df.apply(
        lambda row: (row["center_y"], row["center_x"]) in positions_edge, axis=1
    )

    mm.logger.debug(f"Beads found: {len(positions_all)}")
    mm.logger.debug(f"Beads kept for further analysis: {positions_df['considered_valid'].sum()}")
    mm.logger.debug(
        f"Beads considered for being to close to the edge: {positions_df['considered_lateral_edge'].sum()}"
    )
    mm.logger.debug(
        f"Beads considered for being to close to each other: {positions_df['considered_self_proximity'].sum()}"
    )

    def get_bead_image(row, channel, min_distance):
        return channel[
            :,
            max(0, (row["center_y"] - int(min_distance // 2))) : min(
                channel.shape[1], (row["center_y"] + int(min_distance // 2) + 1)
            ),
            max(0, (row["center_x"] - int(min_distance // 2))) : min(
                channel.shape[2], (row["center_x"] + int(min_distance // 2) + 1)
            ),
        ]

    positions_df["beads"] = positions_df.apply(get_bead_image, axis=1, args=(channel, min_distance))

    return positions_df


def _process_channel(
    channel: np.ndarray,
    sigma: tuple[float, float, float],
    min_bead_distance: float,
    snr_threshold: float,
    fitting_r2_threshold: float,
    intensity_robust_z_score_threshold: float,
    voxel_size_micron: tuple[float, float, float],
) -> pd.DataFrame:
    bead_properties = _find_beads(
        channel=channel,
        sigma=sigma,
        min_distance=min_bead_distance,
    )

    bead_properties = bead_properties.assign(
        considered_intensity_outlier=pd.Series(dtype=bool),
    )

    bead_properties = bead_properties.join(
        bead_properties["beads"].apply(lambda x: pd.Series(_process_bead(x, voxel_size_micron)))
    )
    bead_properties["considered_bad_fit_z"] = bead_properties["fit_r2_z"] < fitting_r2_threshold
    bead_properties["considered_bad_fit_y"] = bead_properties["fit_r2_y"] < fitting_r2_threshold
    bead_properties["considered_bad_fit_x"] = bead_properties["fit_r2_x"] < fitting_r2_threshold

    _calculate_bead_intensity_outliers(
        bead_properties, robust_z_score_threshold=intensity_robust_z_score_threshold
    )

    # We need to invalidate all the bad fits and outliers
    bead_properties["considered_valid"] = bead_properties.apply(
        lambda row: not any(
            [
                row["considered_lateral_edge"],
                row["considered_axial_edge"],
                row["considered_bad_fit_z"],
                row["considered_bad_fit_y"],
                row["considered_bad_fit_x"],
                row["considered_intensity_outlier"],
            ]
        ),
        axis=1,
    )

    return bead_properties


def _process_image(
    image: mm_schema.Image,
    sigma: tuple[float, float, float],
    min_bead_distance: float,
    snr_threshold: float,
    fitting_r2_threshold: float,
    intensity_robust_z_score_threshold: float,
) -> tuple:
    channel_names = [c.name for c in image.channel_series.channels]
    voxel_size_micron = (
        image.voxel_size_z_micron,
        image.voxel_size_y_micron,
        image.voxel_size_x_micron,
    )

    # Get image data and remove the time dimension
    image = image.array_data[0, ...]

    # Some images (e.g. OMX-3D-SIM) may contain negative values.
    image = np.clip(image, a_min=0, a_max=None)

    nr_channels = image.shape[-1]

    bead_properties = []

    for ch in range(nr_channels):
        ch_bead_positions = _process_channel(
            channel=image[..., ch],
            sigma=sigma,
            min_bead_distance=min_bead_distance,
            snr_threshold=snr_threshold,
            fitting_r2_threshold=fitting_r2_threshold,
            intensity_robust_z_score_threshold=intensity_robust_z_score_threshold,
            voxel_size_micron=voxel_size_micron,
        )

        _add_row_index_level(ch_bead_positions, "channel_nr", ch)
        _add_row_index_level(ch_bead_positions, "channel_name", channel_names[ch])
        bead_properties.append(ch_bead_positions)

    return pd.concat(bead_properties)


def _estimate_min_bead_distance(dataset: mm_schema.PSFBeadsDataset) -> float:
    # TODO: get the resolution somewhere or pass it as a metadata
    return dataset.input_parameters.min_lateral_distance_factor


def _generate_center_roi(
    dataset: mm_schema.PSFBeadsDataset,
    positions,
    root_name,
    color,
    stroke_width=1,
):
    rois = []

    for image in dataset.input_data.psf_beads_images:
        image_id = mm.analyses.get_object_id(image) or image.name
        if positions.empty or image_id not in positions.index.get_level_values("image_id"):
            continue
        points = []
        for index, row in positions.xs(image_id, level="image_id").iterrows():
            points.append(
                mm_schema.Point(
                    name=index[positions.index.names[1:].index("bead_id")],
                    z=row["center_z"],
                    y=row["center_y"] + 0.5,  # Rois are centered on the voxel
                    x=row["center_x"] + 0.5,
                    c=index[positions.index.names[1:].index("channel_nr")],
                    stroke_color=mm_schema.Color(
                        r=color[0], g=color[1], b=color[2], alpha=color[3]
                    ),
                    stroke_width=stroke_width,
                )
            )

        if points:
            rois.append(
                mm_schema.Roi(
                    name=f"{root_name}_{image_id}",
                    description=f"{root_name} in image {image_id}",
                    linked_references=image.data_reference,
                    points=points,
                )
            )

    return rois


def _extract_profiles(bead_properties, axis: str) -> pd.DataFrame:
    profile_col_names = [
        f"{axis}_raw",
        f"{axis}_fitted",
    ]
    column_indexes = [i for i in bead_properties.index.names if i != "channel_name"]

    profiles = {}
    for index, row in bead_properties.iterrows():
        if isinstance(index, (list, tuple)):
            index = [index[bead_properties.index.names.index(col_i)] for col_i in column_indexes]

            index_str = "_".join([str(i) for i in index])
        else:
            index_str = str(index)
        for profile_name in profile_col_names:
            profiles[f"{index_str}_{profile_name}"] = pd.Series(row[profile_name])

    bead_properties.drop(columns=profile_col_names, inplace=True)

    return pd.DataFrame(profiles)


def analyse_psf_beads(dataset: mm_schema.PSFBeadsDataset) -> bool:
    mm.analyses.validate_requirements()
    # TODO: Implement Nyquist validation??

    # Containers for input data and input parameters
    images = {}
    voxel_sizes_micron = {}
    min_bead_distance = _estimate_min_bead_distance(dataset)
    snr_threshold = dataset.input_parameters.snr_threshold
    fitting_r2_threshold = dataset.input_parameters.fitting_r2_threshold

    # Containers for output data
    saturated_channels = {}
    bead_properties = []

    # First loop to prepare data and do checks
    images_shape = None
    for image in dataset.input_data.psf_beads_images:
        image_id = mm.analyses.get_object_id(image) or image.name
        images[image_id] = image.array_data[0, ...]

        saturated_channels[image_id] = []

        # Check image shape
        mm.logger.info(f"Checking image {image_id} shape...")
        if len(image.array_data.shape) != 5:
            mm.logger.error(f"Image {image_id} must be 5D")
            return False
        if image.array_data.shape[0] != 1:
            mm.logger.warning(
                f"Image {image_id} must be in TZYXC order and single time-point. Using first time-point."
            )

        # Check all shapes equal
        if images_shape is None:
            images_shape = image.array_data.shape
        elif images_shape != image.array_data.shape:
            mm.logger.error("Not all images have the same dimensions")
            return False

        # Check image saturation
        mm.logger.info(f"Checking image {image_id} saturation...")
        for c in range(image.array_data.shape[-1]):
            if mm_tools.is_saturated(
                channel=image.array_data[..., c],
                threshold=dataset.input_parameters.saturation_threshold,
                detector_bit_depth=dataset.input_parameters.bit_depth,
            ):
                mm.logger.error(f"Image {image_id}: channel {c} is saturated")
                saturated_channels[image_id].append(c)

    if any(len(saturated_channels[name]) for name in saturated_channels):
        mm.logger.error(f"Channels {saturated_channels} are saturated")
        raise mm.SaturationError(f"Channels {saturated_channels} are saturated")

    # Second loop main image analysis
    for image in dataset.input_data.psf_beads_images:
        image_id = mm.analyses.get_object_id(image) or image.name
        mm.logger.info(f"Processing image {image_id}...")
        voxel_sizes_micron[image_id] = (
            image.voxel_size_z_micron,
            image.voxel_size_y_micron,
            image.voxel_size_x_micron,
        )

        image_bead_properties = _process_image(
            image=image,
            sigma=(
                dataset.input_parameters.sigma_z,
                dataset.input_parameters.sigma_y,
                dataset.input_parameters.sigma_x,
            ),
            min_bead_distance=min_bead_distance,
            snr_threshold=snr_threshold,
            fitting_r2_threshold=fitting_r2_threshold,
            intensity_robust_z_score_threshold=dataset.input_parameters.intensity_robust_z_score_threshold,
        )

        mm.logger.info(
            f"Image {image_id} processed."
            f"    {image_bead_properties.considered_valid.sum()} beads considered valid."
            f"    {image_bead_properties.considered_lateral_edge.sum()} beads considered lateral edge."
            f"    {image_bead_properties.considered_self_proximity.sum()} beads considered self proximity."
            f"    {image_bead_properties.considered_axial_edge.sum()} beads considered axial edge."
            f"    {image_bead_properties.considered_intensity_outlier.sum()} beads considered intensity outlier."
            f"    {image_bead_properties.considered_bad_fit_z.sum()} beads considered bad fit in z."
            f"    {image_bead_properties.considered_bad_fit_y.sum()} beads considered bad fit in y."
            f"    {image_bead_properties.considered_bad_fit_x.sum()} beads considered bad fit in x."
        )

        _add_row_index_level(image_bead_properties, "image_id", image_id)
        bead_properties.append(image_bead_properties)

    bead_properties = pd.concat(bead_properties)

    # Before averaging any bead we need to verify that all voxel sizes are equal
    if len(set(voxel_sizes_micron.values())) == 1:
        average_beads_properties = bead_properties.groupby(["channel_nr", "channel_name"]).apply(
            _average_beads
        )
        average_beads_properties = average_beads_properties.join(
            average_beads_properties["average_bead"].apply(
                lambda x: pd.Series(_process_bead(x, voxel_sizes_micron[image_id]))
            )
        )
        average_beads_properties.drop(columns=["considered_axial_edge"], inplace=True)
    else:
        mm.logger.error(
            "Voxel sizes are not equal among images. Skipping average bead calculation."
        )
        average_beads_properties = None

    bead_profiles_z = _extract_profiles(bead_properties, "z")
    bead_profiles_y = _extract_profiles(bead_properties, "y")
    bead_profiles_x = _extract_profiles(bead_properties, "x")

    bead_profiles_z = bead_profiles_z.join(_extract_profiles(average_beads_properties, "z"))
    bead_profiles_y = bead_profiles_y.join(_extract_profiles(average_beads_properties, "y"))
    bead_profiles_x = bead_profiles_x.join(_extract_profiles(average_beads_properties, "x"))

    # TODO: get more metadata from the source images
    if any(isinstance(c, np.ndarray) for c in average_beads_properties["average_bead"]):
        average_bead = mm.analyses.numpy_to_mm_image(
            array=np.expand_dims(
                np.stack(
                    [
                        c
                        for c in average_beads_properties["average_bead"]
                        if isinstance(c, np.ndarray)
                    ],
                    axis=-1,
                ),
                axis=0,
            ),
            name="average_bead",
            description="Average bead image extracted from all the beads considered valid in the dataset.",
            source_images=dataset.input_data.psf_beads_images,
            channel_names=[
                i[average_beads_properties.index.names.index("channel_name")]
                for i in average_beads_properties.index
                if isinstance(average_beads_properties.at[i, "average_bead"], np.ndarray)
            ],
        )
    else:
        average_bead = None
    average_beads_properties.drop("average_bead", axis=1, inplace=True)
    bead_properties.drop("beads", axis=1, inplace=True)

    key_measurements = _generate_key_measurements(
        bead_properties=bead_properties,
        average_bead_properties=average_beads_properties,
    )

    key_measurements = mm_schema.PSFBeadsKeyMeasurements(
        name="psf_beads_key_measurements",
        description="Averaged key measurements for all beads considered valid in the dataset.",
        table_data=key_measurements,
        **key_measurements.to_dict("list"),
    )

    considered_valid_bead_centers = _generate_center_roi(
        dataset=dataset,
        positions=bead_properties[bead_properties.considered_valid],
        root_name="considered_valid_bead_centers",
        color=(0, 255, 0, 100),
        stroke_width=8,
    )
    considered_lateral_edge_bead_centers = _generate_center_roi(
        dataset=dataset,
        positions=bead_properties[bead_properties.considered_lateral_edge],
        root_name="considered_lateral_edge_bead_centers",
        color=(255, 0, 0, 100),
        stroke_width=4,
    )
    considered_self_proximity_bead_centers = _generate_center_roi(
        dataset=dataset,
        positions=bead_properties[bead_properties.considered_self_proximity],
        root_name="considered_self_proximity_bead_centers",
        color=(255, 0, 0, 100),
        stroke_width=4,
    )
    considered_axial_edge_bead_centers = _generate_center_roi(
        dataset=dataset,
        positions=bead_properties[bead_properties.considered_axial_edge],
        root_name="considered_axial_edge_bead_centers",
        color=(0, 0, 255, 100),
        stroke_width=4,
    )
    considered_intensity_outlier_bead_centers = _generate_center_roi(
        dataset=dataset,
        positions=bead_properties[bead_properties.considered_intensity_outlier],
        root_name="considered_intensity_outlier_bead_centers",
        color=(0, 0, 255, 100),
        stroke_width=4,
    )
    considered_bad_fit_z_bead_centers = _generate_center_roi(
        dataset=dataset,
        positions=bead_properties[bead_properties.considered_bad_fit_z],
        root_name="considered_bad_fit_z_bead_centers",
        color=(0, 0, 255, 100),
        stroke_width=4,
    )
    considered_bad_fit_y_bead_centers = _generate_center_roi(
        dataset=dataset,
        positions=bead_properties[bead_properties.considered_bad_fit_y],
        root_name="considered_bad_fit_y_bead_centers",
        color=(0, 0, 255, 100),
        stroke_width=4,
    )
    considered_bad_fit_x_bead_centers = _generate_center_roi(
        dataset=dataset,
        positions=bead_properties[bead_properties.considered_bad_fit_x],
        root_name="considered_bad_fit_x_bead_centers",
        color=(0, 0, 255, 100),
        stroke_width=4,
    )

    bead_properties = mm.analyses.df_to_table(bead_properties.reset_index(), "bead_properties")
    bead_profiles_z = mm.analyses.df_to_table(bead_profiles_z, "bead_profiles_z")
    bead_profiles_y = mm.analyses.df_to_table(bead_profiles_y, "bead_profiles_y")
    bead_profiles_x = mm.analyses.df_to_table(bead_profiles_x, "bead_profiles_x")

    dataset.output = mm_schema.PSFBeadsOutput(
        processing_application="microscopemetrics",
        processing_version=mm.__version__,
        processing_datetime=datetime.now(),
        analyzed_bead_centers=considered_valid_bead_centers,
        considered_bead_centers_lateral_edge=considered_lateral_edge_bead_centers,
        considered_bead_centers_self_proximity=considered_self_proximity_bead_centers,
        considered_bead_centers_axial_edge=considered_axial_edge_bead_centers,
        considered_bead_centers_intensity_outlier=considered_intensity_outlier_bead_centers,
        considered_bead_centers_z_fit_quality=considered_bad_fit_z_bead_centers,
        considered_bead_centers_y_fit_quality=considered_bad_fit_y_bead_centers,
        considered_bead_centers_x_fit_quality=considered_bad_fit_x_bead_centers,
        key_measurements=key_measurements,
        bead_properties=bead_properties,
        bead_profiles_z=bead_profiles_z,
        bead_profiles_y=bead_profiles_y,
        bead_profiles_x=bead_profiles_x,
        average_bead=average_bead,
    )

    dataset.processed = True

    return True


# Calculate 2D FFT
# slice_2d = raw_img[17, ...].reshape([1, n_channels, x_size, y_size])
# fft_2D = fft_2d(slice_2d)

# Calculate 3D FFT
# fft_3D = fft_3d(spots_image)
#
# plt.imshow(np.log(fft_3D[2, :, :, 1]))  # , cmap='hot')
# # plt.imshow(np.log(fft_3D[2, 23, :, :]))  # , cmap='hot')
# plt.show()
#



from datetime import datetime

import microscopemetrics_schema.datamodel as mm_schema
import numpy as np
import pandas as pd
from scipy.ndimage import zoom
from skimage.measure import profile_line

import microscopemetrics as mm


def _generate_profile(
    image: mm_schema.Image,
    roi: mm_schema.Roi,
) -> mm_schema.Table:
    line = roi.lines[0]

    profile = profile_line(
        image=image.array_data[0, int(line.z), :, :, :],  # T  # Z  # Y  # X  # C
        src=(line.x1, line.y1),
        dst=(line.x2, line.y2),
    )

    columns = [
        mm_schema.Column(
            name="position",
            description="Position along the line",
            data_type="float",
            values=list(range(len(profile))),
        )
    ]
    columns.extend(
        [
            mm_schema.Column(
                name=f"intensity_{c}",
                description=f"Intensity value for channel {c}",
                values=profile[..., c].tolist(),
            )
            for c in range(profile.shape[-1])
        ]
    )

    return mm_schema.Table(
        name=f"{image.name}_{roi.name}_profile",
        description=f"Intensity profile of {roi.name} in {image.name}",
        columns=columns,
    )


def _generate_intensity_profiles(
    dataset: mm_schema.UserExperimentDataset,
) -> list[mm_schema.Table]:
    tables = []

    for roi in dataset.input_data.profile_rois:
        if len(roi.linked_references) != 1:
            raise ValueError("ROI must be linked to exactly one image")

        image = next(
            (
                img
                for img in dataset.input_data.user_experiment_images
                if mm.analyses.get_object_id(roi.linked_references[0])
                == mm.analyses.get_object_id(img)
            ),
            None,
        )
        if image is None:
            raise ValueError("ROI must be linked to an image in the dataset")

        tables.append(
            _generate_profile(
                image=image,
                roi=roi,
            )
        )

    return tables


def _generate_orthogonals(
    image: mm_schema.Image,
    roi: mm_schema.Roi,
) -> tuple[list[mm_schema.OrthogonalImage], list[mm_schema.Roi]]:
    try:
        # We are assuming that the ROI is a single point
        coords = (int(roi.points[0].z), int(roi.points[0].y), int(roi.points[0].x))
    except AttributeError as e:
        raise AttributeError("ROI must have points") from e

    # We are assuming that the voxel size is the same for x and y
    z_ratio = image.voxel_size_z_micron / image.voxel_size_x_micron

    orthogonals = [
        mm_schema.OrthogonalImage(
            name=f"{image.name}_XY_{roi.name}",
            description=f"XY orthogonal view of {image.name} at {coords}",
            channel_series=image.channel_series,
            shape_x=image.shape_x,
            shape_y=image.shape_y,
            shape_z=1,
            source_images=image.data_reference,
            source_roi=roi,
            axis="XY",
            array_data=image.array_data[:, coords[0] : (coords[0] + 1), :, :, :],  # dims TZYXC
        ),
        mm_schema.OrthogonalImage(
            name=f"{image.name}_YZ_{roi.name}",
            description=f"YZ orthogonal view of {image.name} at {coords}",
            channel_series=image.channel_series,
            shape_x=int(np.round(image.shape_z * z_ratio)),
            shape_y=image.shape_y,
            shape_z=1,
            source_images=image.data_reference,
            source_roi=roi,
            axis="YZ",
            array_data=zoom(
                input=np.transpose(
                    image.array_data[:, :, :, coords[1] : (coords[1] + 1), :],  # dims TZYXC
                    axes=(0, 3, 2, 1, 4),  # dims TXYZC
                ),
                zoom=(1, 1, 1, z_ratio, 1),
            ),
        ),
        mm_schema.OrthogonalImage(
            name=f"{image.name}_XZ_{roi.name}",
            description=f"XZ orthogonal view of {image.name} at {coords}",
            channel_series=image.channel_series,
            shape_x=image.shape_x,
            shape_y=int(np.round(image.shape_z * z_ratio)),
            shape_z=1,
            source_images=image.data_reference,
            source_roi=roi,
            axis="XZ",
            array_data=zoom(
                input=np.transpose(
                    image.array_data[:, :, coords[2] : (coords[2] + 1), :, :],  # dims TZYXC
                    axes=(0, 2, 1, 3, 4),  # dims TYZXC
                ),
                zoom=(1, 1, z_ratio, 1, 1),
            ),
        ),
    ]

    section_rois = [
        mm_schema.Roi(
            name=f"{roi.name}_XY",
            description=f"XY position of {roi.name}",
            lines=[
                mm_schema.Line(
                    name=f"section_line_XZ_{roi.name}",
                    x1=0,
                    y1=coords[1],
                    x2=orthogonals[0].shape_x,
                    y2=coords[1],
                ),
                mm_schema.Line(
                    name=f"section_line_YZ_{roi.name}",
                    x1=coords[2],
                    y1=0,
                    x2=coords[2],
                    y2=orthogonals[0].shape_y,
                ),
            ],
        ),
        mm_schema.Roi(
            name=f"{roi.name}_YZ",
            description=f"YZ position of {roi.name}",
            lines=[
                mm_schema.Line(
                    name=f"section_line_XY_{roi.name}",
                    x1=coords[0] * z_ratio,
                    y1=0,
                    x2=coords[0] * z_ratio,
                    y2=orthogonals[1].shape_y,
                ),
                mm_schema.Line(
                    name=f"section_line_XZ_{roi.name}",
                    x1=0,
                    y1=coords[1],
                    x2=orthogonals[1].shape_x,
                    y2=coords[1],
                ),
            ],
        ),
        mm_schema.Roi(
            name=f"{roi.name}_XZ",
            description=f"XZ position of {roi.name}",
            lines=[
                mm_schema.Line(
                    name=f"section_line_XY_{roi.name}",
                    x1=0,
                    y1=coords[0] * z_ratio,
                    x2=orthogonals[2].shape_x,
                    y2=coords[0] * z_ratio,
                ),
                mm_schema.Line(
                    name=f"section_line_YZ_{roi.name}",
                    x1=coords[2],
                    y1=0,
                    x2=coords[2],
                    y2=orthogonals[2].shape_y,
                ),
            ],
        ),
    ]

    return orthogonals, section_rois


def _generate_orthogonal_images(
    dataset: mm_schema.UserExperimentDataset,
) -> list[mm_schema.OrthogonalImage]:
    orthogonals = []
    section_rois = []

    for roi in dataset.input_data.orthogonal_rois:
        if len(roi.linked_references) != 1:
            raise ValueError("ROI must be linked to exactly one image")

        image = next(
            (
                img
                for img in dataset.input_data.user_experiment_images
                if mm.analyses.get_object_id(roi.linked_references[0])
                == mm.analyses.get_object_id(img)
            ),
            None,
        )
        if image is None:
            raise ValueError("ROI must be linked to an image in the dataset")

        o, r = _generate_orthogonals(image=image, roi=roi)
        orthogonals.extend(o)
        section_rois.extend(r)

    return orthogonals


def _get_fft_images(
    dataset: mm_schema.UserExperimentDataset,
) -> list[mm_schema.Image]:
    pass


def _get_key_measurements(
    dataset: mm_schema.UserExperimentDataset,
) -> mm_schema.UserExperimentKeyMeasurements:
    pass


def analyse_user_experiment(dataset: mm_schema.UserExperimentDataset) -> bool:
    mm.analyses.validate_requirements()

    # Containers for output data
    saturated_channels = {}

    # First loop to prepare data
    for image in dataset.input_data.user_experiment_images:
        image_id = mm.analyses.get_object_id(image) or image.name

        saturated_channels[image_id] = []

        # Check image shape
        mm.logger.info(f"Checking image {image_id} shape...")
        if len(image.array_data.shape) != 5:
            mm.logger.error(f"Image {image_id} must be 5D")
            return False

        # Check image saturation
        mm.logger.info(f"Checking image {image_id} saturation...")
        for c in range(image.array_data.shape[-1]):
            if mm.analyses.tools.is_saturated(
                channel=image.array_data[..., c],
                threshold=dataset.input_parameters.saturation_threshold,
                detector_bit_depth=dataset.input_parameters.bit_depth,
            ):
                mm.logger.warning(f"Image {image_id}: channel {c} is saturated")
                saturated_channels[image_id].append(c)

    intensity_profiles = _generate_intensity_profiles(dataset)
    orthogonal_images = _generate_orthogonal_images(dataset)
    fft_images = _get_fft_images(dataset)

    dataset.output = mm_schema.UserExperimentOutput(
        processing_application="microscopemetrics",
        processing_version=mm.__version__,
        processing_datetime=datetime.now(),
        intensity_profiles=intensity_profiles,
        orthogonal_images=orthogonal_images,
        fft_images=fft_images,
    )

    dataset.processed = True

    return True



"""
This module maps the sample classes to the analysis functions and the dataset classes.
"""

from collections import namedtuple

from microscopemetrics_schema import datamodel as mm_schema

from microscopemetrics.analyses import field_illumination, psf_beads

Mapping = namedtuple("Mapping", ["sample_class", "analysis_function", "dataset_class"])

MAPPINGS = [
    Mapping(
        mm_schema.FluorescentHomogeneousThinField,
        field_illumination.analyse_field_illumination,
        mm_schema.FieldIlluminationDataset,
    ),
    Mapping(
        mm_schema.FluorescentHomogeneousThickField,
        field_illumination.analyse_field_illumination,
        mm_schema.FieldIlluminationDataset,
    ),
    Mapping(mm_schema.PSFBeads, psf_beads.analyse_psf_beads, mm_schema.PSFBeadsDataset),
]

# TEST = {
#     mm_schema.FieldIlluminationDataset: {
#         "sample_class": [
#             mm_schema.FluorescentHomogeneousThinField,
#             mm_schema.FluorescentHomogeneousThickField,
#         ],
#         "analysis_function": [
#             field_illumination.analyse_field_illumination,
#             field_illumination.analyse_field_illumination,
#         ],
#     },
#     mm_schema.PSFBeadsDataset: {
#         "sample_class": [
#             mm_schema.PSFBeads,
#         ],
#         "analysis_function": [
#             psf_beads.analyse_psf_beads,
#         ],
#     },
# }
#



from datetime import datetime
from itertools import product
from typing import Dict, List, Tuple

import microscopemetrics_schema.datamodel as mm_schema
import numpy as np
import pandas as pd
from numpy import float64, int64, ndarray
from pandas import DataFrame
from scipy.optimize import curve_fit
from scipy.signal import find_peaks
from skimage.transform import hough_line  # hough_line_peaks, probabilistic_hough_line

from microscopemetrics.analyses import logger, numpy_to_mm_image, validate_requirements
from microscopemetrics.analyses.tools import (
    airy_fun,
    compute_distances_matrix,
    compute_spots_properties,
    is_saturated,
    multi_airy_fun,
    segment_image,
)


def _profile_to_columns(profile: ndarray, channel: int) -> List[Dict[str, Dict[str, List[float]]]]:
    table = [{f"raw_profile_ch{channel:02}": {"values": [v.item() for v in profile[0, :]]}}]

    for p in range(1, profile.shape[0]):
        table.append(
            {
                f"fitted_profile_ch{channel:02}_peak{p:03d}": {
                    "values": [v.item() for v in profile[p, :]]
                }
            }
        )

    return table


def _fit(
    profile: ndarray,
    peaks_guess: List[int64],
    amp: int = 4,
    exp: int = 2,
    lower_amp: int = 3,
    upper_amp: int = 5,
    center_tolerance: int = 1,
) -> Tuple[ndarray, ndarray, ndarray]:
    guess = []
    lower_bounds = []
    upper_bounds = []
    for p in peaks_guess:
        guess.append(p)  # peak center
        guess.append(amp)  # peak amplitude
        lower_bounds.append(p - center_tolerance)
        lower_bounds.append(lower_amp)
        upper_bounds.append(p + center_tolerance)
        upper_bounds.append(upper_amp)

    x = np.linspace(0, profile.shape[0], profile.shape[0], endpoint=False)

    popt, pcov = curve_fit(
        f=multi_airy_fun,
        xdata=x,
        ydata=profile,
        p0=guess,
        bounds=(lower_bounds, upper_bounds),
    )

    opt_peaks = popt[::2]
    # opt_amps = [a / 4 for a in popt[1::2]]  # We normalize back the amplitudes to the unity
    opt_amps = popt[1::2]

    fitted_profiles = np.zeros((len(peaks_guess), profile.shape[0]))
    for i, (c, a) in enumerate(zip(opt_peaks, opt_amps)):
        fitted_profiles[i, :] = airy_fun(x, c, a)

    return opt_peaks, opt_amps, fitted_profiles


def _compute_channel_resolution(
    channel: ndarray,
    axis: int,
    prominence_threshold: float,
    measured_band: float,
    do_fitting: bool = True,
    do_angle_refinement: bool = False,
) -> Tuple[ndarray, int64, ndarray, ndarray, ndarray, float64, int]:
    """Computes the resolution on a pattern of lines with increasing separation"""
    # find the most contrasted z-slice
    z_stdev = np.std(channel, axis=(1, 2))
    z_focus = np.argmax(z_stdev)
    focus_slice = channel[z_focus]

    # TODO: verify angle and correct
    if do_angle_refinement:
        # Set a precision of 0.1 degree.
        tested_angles = np.linspace(-np.pi / 2, np.pi / 2, 1800)
        h, theta, d = hough_line(focus_slice, theta=tested_angles)

    # Cut a band of that found peak
    # Best we can do now is just to cut a band in the center
    # We create a profiles along which we average signal
    axis_len = focus_slice.shape[-axis]
    weight_profile = np.zeros(axis_len)
    # Calculates a band of relative width 'image_fraction' to integrate the profile
    weight_profile[
        int((axis_len / 2) - (axis_len * measured_band / 2)) : int(
            (axis_len / 2) + (axis_len * measured_band / 2)
        )
    ] = 1
    profile = np.average(focus_slice, axis=-axis, weights=weight_profile)

    normalized_profile = (profile - np.min(profile)) / np.ptp(profile)

    peak_positions, properties = find_peaks(
        normalized_profile,
        height=0.3,
        distance=2,
        prominence=prominence_threshold / 4,
    )

    # From the properties we are interested in the amplitude
    # peak_heights = [h for h in properties['peak_heights']]
    ray_filtered_peak_pos = []
    ray_filtered_peak_heights = []
    ray_filtered_peak_prominences = []

    for peak, height, prom in zip(
        peak_positions, properties["peak_heights"], properties["prominences"]
    ):
        if (
            prom / height
        ) > prominence_threshold:  # This is calculating the prominence in relation to the local intensity
            ray_filtered_peak_pos.append(peak)
            ray_filtered_peak_heights.append(height)
            ray_filtered_peak_prominences.append(prom)

    peak_positions = ray_filtered_peak_pos
    peak_heights = ray_filtered_peak_heights
    peak_prominences = ray_filtered_peak_prominences

    if do_fitting:
        peak_positions, peak_heights, fitted_profiles = _fit(normalized_profile, peak_positions)
        normalized_profile = np.append(
            np.expand_dims(normalized_profile, 0), fitted_profiles, axis=0
        )

    # Find the closest peaks to return it as a measure of resolution
    peaks_distances = [abs(a - b) for a, b in zip(peak_positions[0:-2], peak_positions[1:-1])]
    if len(peaks_distances) == 0:
        res = None
        res_index = None
    else:
        res = min(peaks_distances)
        res_index = peaks_distances.index(res)

    return (
        normalized_profile,
        z_focus,
        peak_positions,
        peak_heights,
        peak_prominences,
        res,
        res_index,
    )


def _compute_resolution(
    image: ndarray,
    axis: int,
    measured_band: float,
    prominence: float,
    do_angle_refinement: bool = False,
) -> Tuple[
    List[ndarray],
    List[int],
    List[ndarray],
    List[ndarray],
    List[ndarray],
    List[float],
    List[int],
]:
    profiles = []
    z_planes = []
    peaks_positions = []
    peaks_heights = []
    peaks_prominences = []
    resolution_values = []
    resolution_indexes = []

    for c in range(image.shape[-1]):
        (
            prof,
            zp,
            pk_pos,
            pk_heights,
            pk_prominences,
            res,
            res_index,
        ) = _compute_channel_resolution(
            channel=image[..., c],
            axis=axis,
            prominence_threshold=prominence,
            measured_band=measured_band,
            do_angle_refinement=do_angle_refinement,
        )
        profiles.append(prof)
        z_planes.append(zp.item())
        peaks_positions.append(pk_pos)
        peaks_heights.append(pk_heights)
        peaks_prominences.append(pk_prominences)
        resolution_values.append(res.item())
        resolution_indexes.append(res_index)

    return (
        profiles,
        z_planes,
        peaks_positions,
        peaks_heights,
        peaks_prominences,
        resolution_values,
        resolution_indexes,
    )


def analyse_argolight_b(dataset: mm_schema.ArgolightBDataset) -> bool:
    validate_requirements()

    # Check image shape
    logger.info("Checking image shape...")
    image = dataset.input_data.argolight_b_image.data
    if len(image.shape) != 5:
        logger.error("Image must be 5D")
        return False

    # Check image saturation
    logger.info("Checking image saturation...")
    saturated_channels = []
    for c in range(image.shape[-1]):
        if is_saturated(
            channel=image[:, :, :, :, c],
            threshold=dataset.input_parameters.saturation_threshold,
            detector_bit_depth=dataset.input_parameters.bit_depth,
        ):
            logger.error(f"Channel {c} is saturated")
            saturated_channels.append(c)
    if len(saturated_channels):
        logger.error(f"Channels {saturated_channels} are saturated")
        return False

    # Calculating the distance between spots in pixels with a security margin
    min_distance = round(dataset.input_parameters.spots_distance * 0.3)

    # Calculating the maximum tolerated distance in microns for the same spot in a different channels
    max_distance = dataset.input_parameters.spots_distance * 0.4

    labels = segment_image(
        image=image,
        min_distance=min_distance,
        sigma=(
            dataset.input_parameters.sigma_z,
            dataset.input_parameters.sigma_y,
            dataset.input_parameters.sigma_x,
        ),
        method="local_max",
        low_corr_factors=dataset.input_parameters.lower_threshold_correction_factors,
        high_corr_factors=dataset.input_parameters.upper_threshold_correction_factors,
    )

    dataset.output.spots_labels_image = numpy_to_mm_image(  # TODO: this should be a mask
        array=labels,
        name=f"{dataset.input_data.argolight_b_image.name}_spots_labels",
        description=f"Spots labels of {dataset.input_data.argolight_b_image.image_url}",
        image_url=dataset.input_data.argolight_b_image.image_url,
        source_image_url=dataset.input_data.argolight_b_image.image_url,
    )

    spots_properties, spots_positions = compute_spots_properties(
        image=image,
        labels=labels,
        remove_center_cross=dataset.input_parameters.remove_center_cross,
    )

    distances_df = compute_distances_matrix(
        positions=spots_positions,
        max_distance=max_distance,
    )

    properties_kv = []
    properties_ls = []
    distances_kv = []
    spots_centroids = []

    for ch, ch_spot_props in enumerate(spots_properties):
        ch_df = DataFrame()
        ch_properties_kv = {}
        ch_df["channel_nr"] = [ch for _ in ch_spot_props]
        ch_df["mask_labels"] = [p["label"] for p in ch_spot_props]
        ch_df["volume"] = [p["area"] for p in ch_spot_props]
        ch_df["roi_volume_units"] = "VOXEL"
        ch_df["max_intensity"] = [p["max_intensity"] for p in ch_spot_props]
        ch_df["min_intensity"] = [p["min_intensity"] for p in ch_spot_props]
        ch_df["mean_intensity"] = [p["mean_intensity"] for p in ch_spot_props]
        ch_df["integrated_intensity"] = [p["integrated_intensity"] for p in ch_spot_props]
        ch_df["z_weighted_centroid"] = [p["weighted_centroid"][0] for p in ch_spot_props]
        ch_df["y_weighted_centroid"] = [p["weighted_centroid"][1] for p in ch_spot_props]
        ch_df["x_weighted_centroid"] = [p["weighted_centroid"][2] for p in ch_spot_props]
        ch_df["roi_weighted_centroid_units"] = "PIXEL"

        # Key metrics for spots intensities
        ch_properties_kv["channel_nr"] = ch
        ch_properties_kv["nr_of_spots"] = len(ch_df)
        ch_properties_kv["intensity_max_spot"] = ch_df["integrated_intensity"].max().item()
        ch_properties_kv["intensity_max_spot_roi"] = ch_df["integrated_intensity"].argmax().item()
        ch_properties_kv["intensity_min_spot"] = ch_df["integrated_intensity"].min().item()
        ch_properties_kv["intensity_min_spot_roi"] = ch_df["integrated_intensity"].argmin().item()
        ch_properties_kv["mean_intensity"] = ch_df["integrated_intensity"].mean().item()
        ch_properties_kv["median_intensity"] = ch_df["integrated_intensity"].median().item()
        ch_properties_kv["std_mean_intensity"] = ch_df["integrated_intensity"].std().item()
        ch_properties_kv["mad_mean_intensity"] = (
            (ch_df["integrated_intensity"] - ch_df["integrated_intensity"].mean()).abs().mean()
        )
        ch_properties_kv["min_max_intensity_ratio"] = (
            ch_properties_kv["intensity_min_spot"] / ch_properties_kv["intensity_max_spot"]
        )

        properties_ls.append(ch_df)
        properties_kv.append(ch_properties_kv)

        channel_shapes = [
            mm_schema.Point(
                label=str(p["label"]),
                x=p["weighted_centroid"][2].item(),
                y=p["weighted_centroid"][1].item(),
                z=p["weighted_centroid"][0].item(),
                c=ch,
                # TODO: put some color
            )
            for p in ch_spot_props
        ]

        spots_centroids.append(
            mm_schema.Roi(
                label=f"Centroids_ch{ch:02}",
                image=dataset.input_data.argolight_b_image.image_url,
                shapes=channel_shapes,
            )
        )
    properties_kv = {k: [i[k] for i in properties_kv] for k in properties_kv[0]}
    properties_df = pd.concat(properties_ls)

    for a, b in product(
        distances_df.channel_a.explode().unique(),
        distances_df.channel_b.explode().unique(),
    ):
        temp_df = distances_df[(distances_df.channel_a == a) & (distances_df.channel_b == b)]
        a = int(a)
        b = int(b)

        pr_distances_kv = {
            "channel_A": a,
            "channel_B": b,
            "mean_3d_dist": temp_df.dist_3d.mean(),
            "median_3d_dist": temp_df.dist_3d.median(),
            "std_3d_dist": temp_df.dist_3d.std(),
            "mad_3d_dist": (temp_df.dist_3d - temp_df.dist_3d.mean()).abs().mean(),
            "mean_z_dist": temp_df.z_dist.mean(),
            "median_z_dist": temp_df.z_dist.median(),
            "std_z_dist": temp_df.z_dist.std(),
            "mad_z_dist": (temp_df.z_dist - temp_df.z_dist.mean()).abs().mean(),
        }

        distances_kv.append(pr_distances_kv)

    distances_kv = {k: [i[k] for i in distances_kv] for k in distances_kv[0]}

    dataset.output.intensity_key_values = mm_schema.ArgolightBIntensityKeyValues(**properties_kv)

    dataset.output.distance_key_values = mm_schema.ArgolightBDistanceKeyValues(**distances_kv)

    dataset.output.spots_properties = mm_schema.TableAsDict(
        name="spots_properties",
        columns=[
            mm_schema.Column(name=k, values=v)
            for k, v in properties_df.to_dict(orient="list").items()
        ],
    )

    dataset.output.spots_distances = mm_schema.TableAsDict(
        name="spots_distances",
        columns=[
            mm_schema.Column(name=k, values=v)
            for k, v in distances_df.to_dict(orient="list").items()
        ],
    )

    dataset.output.spots_centroids = spots_centroids

    dataset.processing_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    dataset.processed = True

    return True


def analyse_argolight_e(dataset: mm_schema.ArgolightEDataset) -> bool:
    validate_requirements()

    # Check image shape
    pass  # TODO: implement

    # Check image saturation
    pass  # TODO: implement

    # Check for axis value
    pass  # TODO: implement

    image = dataset.input_data.argolight_e_image.data
    image = image[0]  # if there is a time dimension, take the first one
    axis = dataset.input_parameters.orientation_axis
    pixel_size = dataset.input_data.argolight_e_image.voxel_size_x_micron
    measured_band = dataset.input_parameters.measured_band

    (
        profiles,
        z_slices,
        peak_positions,
        peak_heights,
        peak_prominences,
        resolution_values,
        resolution_indexes,
    ) = _compute_resolution(
        image=image,
        axis=axis,
        measured_band=measured_band,
        prominence=dataset.input_parameters.prominence_threshold,
        do_angle_refinement=False,  # TODO: implement angle refinement
    )
    key_values = {
        "channel_nr": [c for c in range(image.shape[-1])],
        "rayleigh_resolution_pixels": resolution_values,
        "rayleigh_resolution_microns": [
            r * pixel_size for r in resolution_values if pixel_size is not None
        ],
        "peak_position_A": [
            peak_positions[ch][ind].item() for ch, ind in enumerate(resolution_indexes)
        ],
        "peak_position_B": [
            peak_positions[ch][ind + 1].item() for ch, ind in enumerate(resolution_indexes)
        ],
        "peak_height_A": [
            peak_heights[ch][ind].item() for ch, ind in enumerate(resolution_indexes)
        ],
        "peak_height_B": [
            peak_heights[ch][ind + 1].item() for ch, ind in enumerate(resolution_indexes)
        ],
        "peak_prominence_A": [
            peak_prominences[ch][ind].item() for ch, ind in enumerate(resolution_indexes)
        ],
        "peak_prominence_B": [
            peak_prominences[ch][ind + 1].item() for ch, ind in enumerate(resolution_indexes)
        ],
        "focus_slice": z_slices,
    }

    out_tables = []
    rois = []

    # Populate tables and rois
    for ch, profile in enumerate(profiles):
        out_tables.append(_profile_to_columns(profile, ch))
        shapes = []
        pos_a, pos_b = (
            key_values["peak_position_A"][ch],
            key_values["peak_position_B"][ch],
        )
        for peak_index, peak in enumerate((pos_a, pos_b)):
            # Measurements are taken at center of pixel, so we add .5 pixel to peak positions
            if axis == 1:  # Y resolution -> horizontal rois
                axis_len = image.shape[2]
                x1_pos = (axis_len / 2) - (axis_len * measured_band / 2)
                y1_pos = peak + 0.5
                x2_pos = (axis_len / 2) + (axis_len * measured_band / 2)
                y2_pos = peak + 0.5
            elif axis == 2:  # X resolution -> vertical rois
                axis_len = image.shape[1]
                y1_pos = (axis_len / 2) - (axis_len * measured_band / 2)
                x1_pos = peak + 0.5
                y2_pos = (axis_len / 2) + (axis_len * measured_band / 2)
                x2_pos = peak + 0.5

            shapes.append(
                mm_schema.Line(
                    label=f"ch{ch:02}_{peak_index}_peak",
                    x1=x1_pos,
                    y1=y1_pos,
                    x2=x2_pos,
                    y2=y2_pos,
                    z=z_slices[ch],
                    c=ch,
                )
            )
        rois.append(
            mm_schema.Roi(
                label=f"ch{ch:02}_peaks",
                image=dataset.input_data.argolight_e_image.image_url,
                shapes=shapes,
            )
        )
    dataset.output.peaks_rois = rois

    dataset.output.key_measurements = mm_schema.ArgolightEKeyValues(**key_values)

    dataset.output.intensity_profiles = [
        mm_schema.TableAsDict(name=f"intensity_profiles_ch{c:02}", columns=out_tables[c])
        for c in range(len(out_tables))
    ]

    dataset.processing_date = datetime.now().strftime("%Y-%m-%d")
    dataset.processed = True

    return True



import numpy as np
import pandas as pd

try:
    from hypothesis import assume
    from hypothesis import strategies as st
except ImportError as e:
    raise ImportError(
        "In order to run the strategies you need to install the test extras. Run `pip install microscopemetrics[test]`."
    ) from e
import microscopemetrics_schema.strategies.analyses as st_mm_analyses_schema
from skimage.exposure import rescale_intensity as skimage_rescale_intensity
from skimage.filters import gaussian as skimage_gaussian
from skimage.util import random_noise as skimage_random_noise

from microscopemetrics.analyses import numpy_to_mm_image


# Strategies for Field Illumination
def _gen_field_illumination_channel(
    y_shape: int,
    x_shape: int,
    y_center_rel_offset: float,
    x_center_rel_offset: float,
    dispersion: float,
    target_min_intensity: float,
    target_max_intensity: float,
    do_noise: bool,
    signal: int,
):
    # Generate the channel as float64
    channel = np.zeros(shape=(y_shape, x_shape), dtype="float64")
    channel[
        int(channel.shape[0] * (0.5 + y_center_rel_offset / 2)),
        int(channel.shape[1] * (0.5 + x_center_rel_offset / 2)),
    ] = 1.0

    channel = skimage_gaussian(
        channel,
        sigma=max(channel.shape) * dispersion,
        mode="constant",
        preserve_range=True,
    )

    # Normalize channel intensity to be between target_min_intensity and target_max_intensity
    # Saturation point is at 1.0 when we rescale later to the target dtype
    channel = skimage_rescale_intensity(
        channel, out_range=(target_min_intensity, target_max_intensity)
    )

    if do_noise:
        # The noise on a 1.0 intensity image is too strong, so we rescale the image to
        # the defined signal and then rescale it back to the target intensity
        channel = channel * signal
        channel = skimage_random_noise(channel, mode="poisson", clip=False)
        channel = channel / signal

    return channel


def _gen_field_illumination_image(
    y_shape: int,
    x_shape: int,
    c_shape: int,
    y_center_rel_offset: list[float],
    x_center_rel_offset: list[float],
    dispersion: list[float],
    target_min_intensity: list[float],
    target_max_intensity: list[float],
    do_noise: bool,
    signal: list[int],
    dtype: np.dtype,
):
    # Generate the image as float64
    image = np.zeros(shape=(y_shape, x_shape, c_shape), dtype="float64")

    for ch in range(c_shape):
        image[:, :, ch] = _gen_field_illumination_channel(
            y_shape=y_shape,
            x_shape=x_shape,
            y_center_rel_offset=y_center_rel_offset[ch],
            x_center_rel_offset=x_center_rel_offset[ch],
            dispersion=dispersion[ch],
            target_min_intensity=target_min_intensity[ch],
            target_max_intensity=target_max_intensity[ch],
            do_noise=do_noise,
            signal=signal[ch],
        )

    # Rescale to the target dtype
    image = np.clip(image, None, 1)
    image = skimage_rescale_intensity(image, in_range=(0.0, 1.0), out_range=dtype)
    image = np.expand_dims(image, (0, 1))

    return image


@st.composite
def st_field_illumination_test_data(
    draw,
    nr_images=st.integers(min_value=1, max_value=3),
    y_image_shape=st.integers(min_value=512, max_value=1024),
    x_image_shape=st.integers(min_value=512, max_value=1024),
    c_image_shape=st.integers(min_value=1, max_value=3),
    dtype=st.sampled_from([np.uint8, np.uint16]),
    signal=st.integers(min_value=100, max_value=1000),
    do_noise=st.just(True),
    target_min_intensity=st.floats(min_value=0.1, max_value=0.45),
    target_max_intensity=st.floats(min_value=0.5, max_value=0.9),
    center_y_relative=st.floats(min_value=-0.8, max_value=0.8),
    center_x_relative=st.floats(min_value=-0.8, max_value=0.8),
    dispersion=st.floats(min_value=0.5, max_value=1.0),
):
    output = {
        "images": [],
        "centers_generated_y_relative": [],
        "centers_generated_x_relative": [],
        "target_min_intensities": [],
        "target_max_intensities": [],
        "dispersions": [],
        "do_noise": [],
        "signals": [],
    }

    y_image_shape = draw(y_image_shape)
    x_image_shape = draw(x_image_shape)

    do_noise = draw(do_noise)

    dtype = draw(dtype)

    for _ in range(draw(nr_images)):
        # We want a different number of channels for each image
        _c_image_shape = draw(c_image_shape)

        image_target_min_intensities = []
        image_target_max_intensities = []
        centers_generated_y_relative = []
        centers_generated_x_relative = []
        image_dispersions = []
        image_signals = []

        for _ in range(_c_image_shape):
            ch_target_min_intensity = draw(target_min_intensity)
            ch_target_max_intensity = draw(target_max_intensity)
            # if the dtype is uint8, the difference between the min and max intensity should be less than 0.5
            # otherwise, with only a few intensity levels, the detection will not be accurate. It is anyway a very
            # unlikely scenario in a real world situation
            if dtype == np.uint8:
                assume((ch_target_max_intensity - ch_target_min_intensity) > 0.7)
            image_target_min_intensities.append(ch_target_min_intensity)
            image_target_max_intensities.append(ch_target_max_intensity)

            ch_y_center_rel_offset = draw(center_y_relative)
            ch_x_center_rel_offset = draw(center_x_relative)
            centers_generated_y_relative.append(ch_y_center_rel_offset)
            centers_generated_x_relative.append(ch_x_center_rel_offset)

            ch_dispersion = draw(dispersion)
            image_dispersions.append(ch_dispersion)

            ch_signal = draw(signal)
            image_signals.append(ch_signal)

        image = _gen_field_illumination_image(
            y_shape=y_image_shape,
            x_shape=x_image_shape,
            c_shape=_c_image_shape,
            y_center_rel_offset=centers_generated_y_relative,
            x_center_rel_offset=centers_generated_x_relative,
            dispersion=image_dispersions,
            target_min_intensity=image_target_min_intensities,
            target_max_intensity=image_target_max_intensities,
            do_noise=do_noise,
            signal=image_signals,
            dtype=dtype,
        )

        output["images"].append(image)
        output["centers_generated_y_relative"].append(centers_generated_y_relative)
        output["centers_generated_x_relative"].append(centers_generated_x_relative)
        output["target_min_intensities"].append(image_target_min_intensities)
        output["target_max_intensities"].append(image_target_max_intensities)
        output["dispersions"].append(image_dispersions)
        output["do_noise"].append(do_noise)
        output["signals"].append(image_signals)

    return output


@st.composite
def st_field_illumination_dataset(
    draw,
    unprocessed_dataset=st_mm_analyses_schema.st_mm_field_illumination_unprocessed_dataset(),
    test_data=st_field_illumination_test_data(),
):
    test_data = draw(test_data)
    field_illumination_unprocessed_dataset = draw(unprocessed_dataset)

    field_illumination_unprocessed_dataset.input_data.field_illumination_images = [
        numpy_to_mm_image(
            array=image,
            name=f"FI_image_{i}",
            channel_names=[f"Channel_{i}{j}" for j in range(image.shape[-1])],
        )
        for i, image in enumerate(test_data.pop("images"))
    ]

    # Setting the bit depth to the data type of the image
    image_dtype = {
        a.array_data.dtype
        for a in field_illumination_unprocessed_dataset.input_data.field_illumination_images
    }
    if len(image_dtype) != 1:
        raise ValueError("All images should have the same data type")
    image_dtype = image_dtype.pop()
    if np.issubdtype(image_dtype, np.integer):
        field_illumination_unprocessed_dataset.input_parameters.bit_depth = np.iinfo(
            image_dtype
        ).bits
    elif np.issubdtype(image_dtype, np.floating):
        field_illumination_unprocessed_dataset.input_parameters.bit_depth = np.finfo(
            image_dtype
        ).bits
    else:
        field_illumination_unprocessed_dataset.input_parameters.bit_depth = None

    return {
        "unprocessed_dataset": field_illumination_unprocessed_dataset,
        "expected_output": test_data,
    }


@st.composite
def st_field_illumination_table(
    draw,
    nr_rows=st.integers(min_value=1, max_value=50),
):
    nr_rows = draw(nr_rows)
    columns = [
        "bottom_center_intensity_mean",
        "bottom_center_intensity_ratio",
        "channel_nr",
    ]
    table = []
    for _ in range(nr_rows):
        dataset = draw(st_field_illumination_dataset())["unprocessed_dataset"]
        dataset.run()
        if dataset.processed:
            key_values = {col: getattr(dataset.output.key_values, col) for col in columns}
        else:
            continue
        table.append(key_values)

    table = [pd.DataFrame(d) for d in table]

    return pd.concat(table, ignore_index=True)



# Main analyses module defining the sample superclass
from datetime import datetime
from typing import Dict, List, Union

import microscopemetrics_schema.datamodel as mm_schema
import numpy as np
import pandas as pd

from microscopemetrics import logger


# TODO: This function is getting the id from OMERO. It should be more general
def get_object_id(
    objects: Union[mm_schema.MetricsObject, List[mm_schema.MetricsObject]]
) -> Union[str, List[str]]:
    """Get the object id of a metrics object or a list of metrics objects"""
    if isinstance(objects, list):
        return [get_object_id(obj) for obj in objects]
    if not isinstance(objects, mm_schema.MetricsObject):
        raise ValueError("Input should be a metrics object or a list of metrics objects")
    if objects.data_reference:
        try:
            return objects.data_reference.omero_object_id
        except AttributeError:
            logger.warning(f"Object {objects.name} does not have an object id")
            return None


def numpy_to_mm_image(
    array: np.ndarray,
    name: str = None,
    description: str = None,
    source_images: List[mm_schema.Image] = None,
    acquisition_datetime: str = None,
    channel_names: List[str] = None,
    channel_descriptions: List[str] = None,
    excitation_wavelengths_nm: List[float] = None,
    emission_wavelengths_nm: List[float] = None,
) -> mm_schema.Image:
    """Converts a numpy array with dimensions order tzyxc to an image by reference (not inlined)"""
    if array.ndim == 5:
        shape_t, shape_z, shape_y, shape_x, shape_c = array.shape
    elif array.ndim == 2:
        shape_y, shape_x = array.shape
        shape_t, shape_z, shape_c = 1, 1, 1
        array = array.reshape((1, 1, shape_y, shape_x, 1))
    else:
        raise NotImplementedError(
            f"Array of dimension {array.ndim} is not supported by this function. Image has to have either 5 or 2 dimensions"
        )

    if source_images:
        source_images_refs = [
            i.data_reference for i in source_images if i.data_reference is not None
        ]
    else:
        source_images_refs = None

    if acquisition_datetime is None:
        if source_images is not None and len(source_images) == 1:
            acquisition_datetime = source_images[0].acquisition_datetime
        else:
            acquisition_datetime = datetime.now().isoformat()

    if channel_names is not None and len(channel_names) != shape_c:
        raise ValueError(
            "The number of channel names should be equal to the number of channels in the image"
        )
    if channel_descriptions is not None and len(channel_descriptions) != shape_c:
        raise ValueError(
            "The number of channel descriptions should be equal to the number of channels in the image"
        )
    if excitation_wavelengths_nm is not None and len(excitation_wavelengths_nm) != shape_c:
        raise ValueError(
            "The number of excitation wavelengths should be equal to the number of channels in the image"
        )
    if emission_wavelengths_nm is not None and len(emission_wavelengths_nm) != shape_c:
        raise ValueError(
            "The number of emission wavelengths should be equal to the number of channels in the image"
        )

    channels = []
    for i in range(shape_c):
        channel = mm_schema.Channel(
            name=channel_names[i] if channel_names is not None else str(i),
            description=(channel_descriptions[i] if channel_descriptions is not None else None),
            excitation_wavelength_nm=(
                excitation_wavelengths_nm[i] if excitation_wavelengths_nm is not None else None
            ),
            emission_wavelength_nm=(
                emission_wavelengths_nm[i] if emission_wavelengths_nm is not None else None
            ),
        )
        channels.append(channel)

    return mm_schema.Image(
        name=name,
        description=description,
        source_images=source_images_refs,
        array_data=array,
        shape_t=shape_t,
        shape_z=shape_z,
        shape_y=shape_y,
        shape_x=shape_x,
        shape_c=shape_c,
        acquisition_datetime=acquisition_datetime,
        channel_series=mm_schema.ChannelSeries(channels=channels),
    )


def _create_table(
    data: Union[dict[str, list], pd.DataFrame],
    name: str,
    description: str = None,
    column_descriptions: dict[str, str] = None,
) -> mm_schema.Table:
    if len(data) == 0:
        logger.error(f"Table {name} could not created as there is no data")
        return None

    # TODO: Add values to columns
    if isinstance(data, dict):
        columns = [mm_schema.Column(name=n, values=v) for n, v in data.items()]
    elif isinstance(data, pd.DataFrame):
        columns = [mm_schema.Column(name=n, values=data[n].tolist()) for n in data.columns]
    else:
        raise ValueError("Data should be either a dictionary or a pandas dataframe")

    if column_descriptions is not None:
        for column in columns:
            try:
                column.description = column_descriptions[column.name]
            except KeyError:
                logger.warning(f"Column {column.name} does not have a description")

    return mm_schema.Table(
        name=name,
        description=description,
        columns=columns,
        table_data=data,
    )


def dict_to_table(
    dictionary: dict[str, list],
    name: str,
    description: str = None,
    column_descriptions: dict[str, str] = None,
) -> mm_schema.Table:
    """Converts a dictionary to a microscope-metrics table"""
    if any(len(dictionary[k]) != len(dictionary[list(dictionary)[0]]) for k in dictionary):
        logger.error(f"Table {name} could not created as the columns have different lengths")
        raise ValueError(
            f"Table {name} could not be created. All columns should have the same length"
        )

    if not all(dictionary[k] for k in dictionary):
        logger.warning(f"Table {name} was created empty. All the column values are empty")

    return _create_table(
        name=name,
        description=description,
        column_descriptions=column_descriptions,
        data=dictionary,
    )


def df_to_table(
    dataframe: pd.DataFrame,
    name: str,
    description: str = None,
    column_descriptions: Dict[str, str] = None,
) -> mm_schema.Table:
    """Converts a df to a microscope-metrics table"""
    return _create_table(
        name=name,
        description=description,
        column_descriptions=column_descriptions,
        data=dataframe,
    )


def validate_requirements() -> bool:
    logger.info("Validating requirements...")
    # TODO: check image dimensions/shape
    return True



__version__ = "0.3.0"

import logging

logger = logging.getLogger(__name__)


class SaturationError(Exception):
    pass


class FittingError(Exception):
    pass


class AnalysisError(Exception):
    pass



from datetime import datetime
from math import hypot
from typing import Dict, Tuple

import microscopemetrics_schema.datamodel as mm_schema
import numpy as np
import pandas as pd
import scipy
from skimage.exposure import rescale_intensity
from skimage.filters import gaussian
from skimage.measure import regionprops

import microscopemetrics as mm
from microscopemetrics.analyses import tools as mm_tools


def _get_center_region_mask(channel: np.ndarray, fraction: float = 0.1) -> np.ndarray:
    """
    Compute the mask of the center region of a channel
    Parameters
    ----------
    channel : np.array.
        image on a 2d np.ndarray format.
    Returns
    -------
    center_region_mask : np.ndarray
        2d bool np.ndarray representing the center region of the chosen channel.
    """
    channel = rescale_intensity(channel, in_range=(0, channel.max()), out_range=(0.0, 1.0))
    mask = channel > (1 - fraction)
    return mask


def _channel_line_profile(
    channel: np.ndarray, start: Tuple[int, int], end: Tuple[int, int], profile_size: int
) -> np.ndarray:
    """
    Compute the intensity profile along a line between x0-y0 and x1-y1 using cubic interpolation. The mode used is
    'nearest' to avoid edge artifacts.
    Parameters
    ----------
    channel : np.array.
        image on a 2d np.ndarray format.
    start : (int, int)
        coordinates of the starting pixel
    end : (int, int)
        coordinates of the ending pixel
    Returns
    -------
    line_pixel_values : np.ndarray
        1d np.ndarray representing the values of the chosen line of pixels.
    """
    x, y = np.linspace(start[0], end[0], profile_size), np.linspace(start[1], end[1], profile_size)

    return scipy.ndimage.map_coordinates(
        input=channel,
        coordinates=np.vstack((x, y)),
        mode="nearest",
    )


def _image_line_profile(image: np.ndarray, profile_size: int) -> Dict[str, list]:
    """
    Compute the intensity profile along a line between x0-y0 and x1-y1
    Parameters
    ----------
    image : np.ndarray.
        image on a 5d np.ndarray format tzyxc.
    profile_size : int
        size of the intensity profile.
    Returns
    -------
    line_pixel_values : dict
        dictionary representing the values of the chosen line of pixels for each channel.
    """
    profile_coordinates = {
        "leftTop_to_rightBottom": ((0, 0), (image.shape[-2], image.shape[-3])),
        "leftBottom_to_rightTop": ((0, image.shape[-3]), (image.shape[-2], 0)),
        "center_horizontal": (
            (0, image.shape[-3] // 2),
            (image.shape[-2], image.shape[-3] // 2),
        ),
        "center_vertical": (
            (image.shape[-2] // 2, 0),
            (image.shape[-2] // 2, image.shape[-3]),
        ),
    }
    output = {}
    for profile_name, (start, end) in profile_coordinates.items():
        profiles = np.zeros((image.shape[-1], 255))
        for c in range(image.shape[-1]):
            profiles[c, :] = _channel_line_profile(
                np.squeeze(image[0, 0, :, :, c]), start, end, profile_size
            )
        output = output | {
            f"ch{c:02}_{profile_name}": profiles[c].tolist() for c in range(image.shape[-1])
        }

    return output


def _line_profile_shapes(image: np.ndarray):
    stroke_color = {"r": 0, "g": 0, "b": 255, "alpha": 200}
    return [
        mm_schema.Line(
            name="leftTop_to_rightBottom",
            x1=0,
            y1=0,
            x2=image.shape[-2],
            y2=image.shape[-3],
            stroke_color=stroke_color,
        ),
        mm_schema.Line(
            name="leftBottom_to_rightTop",
            x1=0,
            y1=image.shape[-3],
            x2=image.shape[-2],
            y2=0,
            stroke_color=stroke_color,
        ),
        mm_schema.Line(
            name="center_horizontal",
            x1=0,
            y1=image.shape[-3] // 2,
            x2=image.shape[-2],
            y2=image.shape[-3] // 2,
            stroke_color=stroke_color,
        ),
        mm_schema.Line(
            name="center_vertical",
            x1=image.shape[-2] // 2,
            y1=0,
            x2=image.shape[-2] // 2,
            y2=image.shape[-3],
            stroke_color=stroke_color,
        ),
    ]


def _c_shape(name, x, y, size, s_col):
    return mm_schema.Rectangle(name=name, x=x, y=y, w=size, h=size, stroke_color=s_col)


def _corner_shapes(image: np.ndarray, corner_fraction: float):
    cfp = int(corner_fraction * (image.shape[-3] + image.shape[-2]) / 2)
    cr_y = int((image.shape[-3] - cfp) / 2)
    cr_x = int((image.shape[-2] - cfp) / 2)
    stroke_color = {"r": 0, "g": 255, "b": 0, "alpha": 200}

    return [
        _c_shape("top_left", x=0, y=0, size=cfp, s_col=stroke_color),
        _c_shape("top_center", x=cr_x, y=0, size=cfp, s_col=stroke_color),
        _c_shape("top_right", x=image.shape[-2] - cfp, y=0, size=cfp, s_col=stroke_color),
        _c_shape("middle_left", x=0, y=cr_y, size=cfp, s_col=stroke_color),
        _c_shape("middle_center", x=cr_x, y=cr_y, size=cfp, s_col=stroke_color),
        _c_shape(
            "middle_right",
            x=image.shape[-2] - cfp,
            y=cr_y,
            size=cfp,
            s_col=stroke_color,
        ),
        _c_shape("bottom_left", x=0, y=image.shape[-3] - cfp, size=cfp, s_col=stroke_color),
        _c_shape(
            "bottom_center",
            x=cr_x,
            y=image.shape[-3] - cfp,
            size=cfp,
            s_col=stroke_color,
        ),
        _c_shape(
            "bottom_right",
            y=image.shape[-3] - cfp,
            x=image.shape[-2] - cfp,
            size=cfp,
            s_col=stroke_color,
        ),
    ]


def _channel_max_intensity_properties(
    channel: np.ndarray,
    sigma: float,
) -> dict:
    """
    Compute the maximum intensity properties of a channel
    """
    if sigma is not None:
        proc_channel = gaussian(image=channel, sigma=sigma, preserve_range=True, channel_axis=None)
    else:
        proc_channel = channel

    # When images are very flat, the max intensity region is always detected in the center. We need to stretch the
    # intensity of the image to detect the actual center. We loop over a list of number of bins to get a central region
    # that is not too big. We consider 0.25 as the maximum fraction of the image that can be considered as the center.
    center_region_area_fraction = 1
    center_region_intensity_fraction = None
    properties = None
    for n_bins in [11, 21, 51, 101, 201, 501]:
        if center_region_area_fraction < 0.2:
            break
        rescaled_channel = rescale_intensity(
            # We scale in 1 value more than the bins we want to achieve.
            # Like that the top value is not included in the last bin.
            proc_channel.astype(float),
            in_range=(0, proc_channel.max()),
            out_range=(0, n_bins),
        )
        labels_channel = rescaled_channel.astype(int)
        properties = regionprops(labels_channel, proc_channel)
        center_region_area_fraction = properties[-2].area / (channel.shape[0] * channel.shape[1])
        center_region_intensity_fraction = 1 / (n_bins - 1)

    # Fitting the intensity profile to a gaussian
    _, _, _, (_, _, center_fitted_y, _) = mm_tools.fit_gaussian(np.max(channel, axis=1))
    _, _, _, (_, _, center_fitted_x, _) = mm_tools.fit_gaussian(np.max(channel, axis=0))

    return {
        "center_region_intensity_fraction": center_region_intensity_fraction,
        "center_region_area_fraction": center_region_area_fraction,
        "center_of_mass_y": properties[-2].centroid[0],
        "center_of_mass_y_relative": properties[-2].centroid[0] / (channel.shape[0] / 2) - 1,
        "center_of_mass_x": properties[-2].centroid[1],
        "center_of_mass_x_relative": properties[-2].centroid[1] / (channel.shape[1] / 2) - 1,
        "center_of_mass_distance_relative": hypot(
            properties[-2].centroid[0] / (channel.shape[0] / 2) - 1,
            properties[-2].centroid[1] / (channel.shape[1] / 2) - 1,
        ),
        "center_geometric_y": properties[-2].centroid[0],
        "center_geometric_y_relative": properties[-2].centroid[0] / (channel.shape[0] / 2) - 1,
        "center_geometric_x": properties[-2].centroid[1],
        "center_geometric_x_relative": properties[-2].centroid[1] / (channel.shape[1] / 2) - 1,
        "center_geometric_distance_relative": hypot(
            properties[-2].centroid[0] / (channel.shape[0] / 2) - 1,
            properties[-2].centroid[1] / (channel.shape[1] / 2) - 1,
        ),
        "center_fitted_y": center_fitted_y,
        "center_fitted_y_relative": center_fitted_y / (channel.shape[0] / 2) - 1,
        "center_fitted_x": center_fitted_x,
        "center_fitted_x_relative": center_fitted_x / (channel.shape[1] / 2) - 1,
        "center_fitted_distance_relative": hypot(
            center_fitted_y / (channel.shape[0] / 2) - 1,
            center_fitted_x / (channel.shape[1] / 2) - 1,
        ),
        "max_intensity": properties[-2].intensity_max,
        "max_intensity_pos_y": properties[-1].centroid[0],
        "max_intensity_pos_y_relative": properties[-1].centroid[0] / (channel.shape[0] / 2) - 1,
        "max_intensity_pos_x": properties[-1].centroid[1],
        "max_intensity_pos_x_relative": properties[-1].centroid[1] / (channel.shape[1] / 2) - 1,
        "max_intensity_distance_relative": hypot(
            properties[-1].centroid[0] / (channel.shape[0] / 2) - 1,
            properties[-1].centroid[1] / (channel.shape[1] / 2) - 1,
        ),
    }


def _channel_corner_properties(channel: np.ndarray, corner_fraction: float) -> dict:
    max_intensity = np.max(channel)

    # Calculate the corner fraction in pixels (cfp) of the image size
    # to use as the corner size and the center range (cr)
    cfp = int(corner_fraction * (channel.shape[0] + channel.shape[1]) / 2)
    cr_y = int((channel.shape[0] - cfp) / 2)
    cr_x = int((channel.shape[1] - cfp) / 2)

    return {
        "top_left_intensity_mean": np.mean(channel[0:cfp, 0:cfp]),
        "top_left_intensity_ratio": np.mean(channel[0:cfp, 0:cfp]) / max_intensity,
        "top_center_intensity_mean": np.mean(channel[0:cfp, cr_x:-cr_x]),
        "top_center_intensity_ratio": np.mean(channel[0:cfp, cr_x:-cr_x]) / max_intensity,
        "top_right_intensity_mean": np.mean(channel[0:cfp, -cfp:-1]),
        "top_right_intensity_ratio": np.mean(channel[0:cfp, -cfp:-1]) / max_intensity,
        "middle_left_intensity_mean": np.mean(channel[cr_y:-cr_y, 0:cfp]),
        "middle_left_intensity_ratio": np.mean(channel[cr_y:-cr_y, 0:cfp]) / max_intensity,
        "middle_center_intensity_mean": np.mean(channel[cr_y:-cr_y, cr_x:-cr_x]),
        "middle_center_intensity_ratio": np.mean(channel[cr_y:-cr_y, cr_x:-cr_x]) / max_intensity,
        "middle_right_intensity_mean": np.mean(channel[cr_y:-cr_y, -cfp:-1]),
        "middle_right_intensity_ratio": np.mean(channel[cr_y:-cr_y, -cfp:-1]) / max_intensity,
        "bottom_left_intensity_mean": np.mean(channel[-cfp:-1, 0:cfp]),
        "bottom_left_intensity_ratio": np.mean(channel[-cfp:-1, 0:cfp]) / max_intensity,
        "bottom_center_intensity_mean": np.mean(channel[-cfp:-1, cr_x:-cr_x]),
        "bottom_center_intensity_ratio": np.mean(channel[-cfp:-1, cr_x:-cr_x]) / max_intensity,
        "bottom_right_intensity_mean": np.mean(channel[-cfp:-1, -cfp:-1]),
        "bottom_right_intensity_ratio": np.mean(channel[-cfp:-1, -cfp:-1]) / max_intensity,
    }


def _image_properties(images: list[mm_schema.Image], corner_fraction: float, sigma: float):
    """
    given FI input images, this function return intensities for the corner and central regions
    and their ratio over the maximum intensity value of the array.
    Parameters
    ----------
    images : list[mm_schema.Image]
        input images from the field illumination dataset.
    Returns
    -------
    profiles_statistics : pandas.DataFrame
        Dictionary showing the intensity values of the different regions and
        their ratio over the maximum intensity value of the array.
        Dictionary values will be lists in case of multiple channels.
    """
    properties = pd.DataFrame()
    for image in images:
        # For the analysis we are using only the first z and time-point
        image_data = image.array_data[0, 0, :, :, :]

        im_properties = pd.DataFrame()
        for c in range(image_data.shape[-1]):
            ch_properties = pd.DataFrame(
                columns=[
                    "image_name",
                    "image_id",
                    "channel_name",
                    "channel_nr",
                    "channel_id",
                ]
            )
            ch_properties.loc[0] = [
                image.name,
                mm.analyses.get_object_id(image),
                image.channel_series.channels[c].name,
                c,
                mm.analyses.get_object_id(image.channel_series.channels[c]),
            ]
            ch_properties = ch_properties.join(
                pd.DataFrame(
                    _channel_max_intensity_properties(image_data[:, :, c], sigma),
                    index=[0],
                )
            )
            ch_properties = ch_properties.join(
                pd.DataFrame(
                    _channel_corner_properties(image_data[:, :, c], corner_fraction),
                    index=[0],
                )
            )
            im_properties = pd.concat([im_properties, ch_properties], axis=0)

        properties = pd.concat([properties, im_properties], axis=0)

    return properties


def analyse_field_illumination(dataset: mm_schema.FieldIlluminationDataset) -> bool:
    mm.analyses.validate_requirements()

    channel_names = []
    for image in dataset.input_data.field_illumination_images:
        # We want to verify that the input images all have different channel names
        # As it does not make sense to average file illumination between images from the same channel
        if image.channel_series is not None:
            mm.logger.info("Checking duplicate channel names...")
            for channel in image.channel_series.channels:
                if channel.name in channel_names:
                    mm.logger.error(
                        f"Channel name {channel.name} is not unique. "
                        "We cannot average field illumination between images from the same channel."
                    )
                    return False
                channel_names.append(channel.name)

        # Check image shape
        mm.logger.info("Checking image shape...")
        if len(image.array_data.shape) != 5:
            mm.logger.error("Image must be 5D")
            return False
        if image.array_data.shape[0] != 1 or image.array_data.shape[1] != 1:
            mm.logger.warning(
                "Image must be in TZYXC order, single z and single time-point. Using first z and time-point."
            )

        # Check image saturation
        mm.logger.info("Checking image saturation...")
        saturated_channels = []
        for c in range(image.array_data.shape[-1]):
            if mm_tools.is_saturated(
                channel=image.array_data[..., c],
                threshold=dataset.input_parameters.saturation_threshold,
                detector_bit_depth=dataset.input_parameters.bit_depth,
            ):
                mm.logger.error(f"Channel {c} is saturated")
                saturated_channels.append(c)
        if len(saturated_channels):
            mm.logger.error(f"Channels {saturated_channels} are saturated")
            raise mm.SaturationError(f"Channels {saturated_channels} are saturated")

    key_measurements = _image_properties(
        images=dataset.input_data.field_illumination_images,
        corner_fraction=dataset.input_parameters.corner_fraction,
        sigma=dataset.input_parameters.sigma,
    )

    key_measurements = mm_schema.FieldIlluminationKeyMeasurements(
        name="field_illumination_key_measurements",
        description="Key measurements of the field illumination channels",
        table_data=key_measurements,
        **key_measurements.to_dict(orient="list"),
    )

    intensity_profiles = [
        mm.analyses.dict_to_table(
            dictionary=_image_line_profile(image.array_data, profile_size=255),
            name=f"{image.name}_intensity_profiles",
            description=f"Intensity profiles of {image.name}",
        )
        for image in dataset.input_data.field_illumination_images
    ]

    roi_profiles = [
        mm_schema.Roi(
            name="Profile ROIs",
            description="ROIs used to compute the intensity profiles",
            linked_references=image.data_reference,
            lines=_line_profile_shapes(image.array_data),
        )
        for image in dataset.input_data.field_illumination_images
    ]

    roi_corners = [
        mm_schema.Roi(
            name="Corner ROIs",
            description="ROIs used to compute the corner intensities",
            linked_references=image.data_reference,
            rectangles=_corner_shapes(image.array_data, dataset.input_parameters.corner_fraction),
        )
        for image in dataset.input_data.field_illumination_images
    ]

    roi_centers_of_mass = [
        mm_schema.Roi(
            name="Centers of mass ROIs",
            description="Point ROI marking the centroids of the max intensity regions",
            linked_references=image.data_reference,
            points=[
                mm_schema.Point(
                    name=f"ch{c:02}_center",
                    y=key_measurements.center_of_mass_y[
                        key_measurements.channel_name.index(image.channel_series.channels[c].name)
                    ],
                    x=key_measurements.center_of_mass_x[
                        key_measurements.channel_name.index(image.channel_series.channels[c].name)
                    ],
                    c=c,
                    stroke_color={"r": 255, "g": 0, "b": 0, "alpha": 200},
                    fill_color={"r": 255, "g": 0, "b": 0, "alpha": 200},
                    stroke_width=5,
                )
                for c in range(image.array_data.shape[-1])
            ],
        )
        for image in dataset.input_data.field_illumination_images
    ]

    roi_centers_geometric = [
        mm_schema.Roi(
            name="Geometric centers ROIs",
            description="Point ROI marking the centroids of the max intensity regions",
            linked_references=image.data_reference,
            points=[
                mm_schema.Point(
                    name=f"ch{c:02}_center",
                    y=key_measurements.center_geometric_y[
                        key_measurements.channel_name.index(image.channel_series.channels[c].name)
                    ],
                    x=key_measurements.center_geometric_x[
                        key_measurements.channel_name.index(image.channel_series.channels[c].name)
                    ],
                    c=c,
                    stroke_color={"r": 255, "g": 0, "b": 0, "alpha": 200},
                    fill_color={"r": 255, "g": 0, "b": 0, "alpha": 200},
                    stroke_width=5,
                )
                for c in range(image.array_data.shape[-1])
            ],
        )
        for image in dataset.input_data.field_illumination_images
    ]

    roi_centers_fitted = [
        mm_schema.Roi(
            name="Fitted centers ROIs",
            description="Point ROI marking the centroids of the max intensity regions",
            linked_references=image.data_reference,
            points=[
                mm_schema.Point(
                    name=f"ch{c:02}_center",
                    y=key_measurements.center_fitted_y[
                        key_measurements.channel_name.index(image.channel_series.channels[c].name)
                    ],
                    x=key_measurements.center_fitted_x[
                        key_measurements.channel_name.index(image.channel_series.channels[c].name)
                    ],
                    c=c,
                    stroke_color={"r": 255, "g": 0, "b": 0, "alpha": 200},
                    fill_color={"r": 255, "g": 0, "b": 0, "alpha": 200},
                    stroke_width=5,
                )
                for c in range(image.array_data.shape[-1])
            ],
        )
        for image in dataset.input_data.field_illumination_images
    ]

    roi_centers_max_intensity = [
        mm_schema.Roi(
            name="Max intensity ROIs",
            description="Point ROI marking the centroids of the max intensity regions",
            linked_references=image.data_reference,
            points=[
                mm_schema.Point(
                    name=f"ch{c:02}_center",
                    y=key_measurements.max_intensity_pos_y[
                        key_measurements.channel_name.index(image.channel_series.channels[c].name)
                    ],
                    x=key_measurements.max_intensity_pos_x[
                        key_measurements.channel_name.index(image.channel_series.channels[c].name)
                    ],
                    c=c,
                    stroke_color={"r": 255, "g": 0, "b": 0, "alpha": 200},
                    fill_color={"r": 255, "g": 0, "b": 0, "alpha": 200},
                    stroke_width=5,
                )
                for c in range(image.array_data.shape[-1])
            ],
        )
        for image in dataset.input_data.field_illumination_images
    ]

    roi_center_region = [
        mm_schema.Roi(
            name="Center region ROIs",
            description="Mask ROIs marking the center region of the image",
            linked_references=image.data_reference,
            masks=[
                mm_schema.Mask(
                    name=f"ch{c:02}_center_region",
                    x=0,
                    y=0,
                    c=c,
                    mask=mm.analyses.numpy_to_mm_image(
                        array=_get_center_region_mask(image.array_data[0, 0, :, :, c])
                    ),
                    fill_color={"r": 255, "g": 200, "b": 60, "alpha": 128},
                )
                for c in range(image.array_data.shape[-1])
            ],
        )
        for image in dataset.input_data.field_illumination_images
    ]

    dataset.output = mm_schema.FieldIlluminationOutput(
        processing_application="microscopemetrics",
        processing_version=mm.__version__,
        processing_datetime=datetime.now(),
        key_measurements=key_measurements,
        intensity_profiles=intensity_profiles,
        roi_profiles=roi_profiles,
        roi_corners=roi_corners,
        roi_centers_of_mass=roi_centers_of_mass,
        roi_centers_geometric=roi_centers_geometric,
        roi_centers_fitted=roi_centers_fitted,
        roi_centers_max_intensity=roi_centers_max_intensity,
        roi_center_region=roi_center_region,
    )

    dataset.processed = True

    return True



import random

import numpy as np
import pandas as pd

try:
    from hypothesis import assume
    from hypothesis import strategies as st
except ImportError as e:
    raise ImportError(
        "In order to run the strategies you need to install the test extras. Run `pip install microscopemetrics[test]`."
    ) from e
import microscopemetrics_schema.strategies.analyses as st_mm_analyses_schema
from skimage.exposure import rescale_intensity as skimage_rescale_intensity
from skimage.filters import gaussian as skimage_gaussian
from skimage.util import random_noise as skimage_random_noise

from microscopemetrics.analyses import numpy_to_mm_image


# Strategies for PSF beads
def _gen_psf_beads_channel():
    pass


def _gen_psf_beads_image(
    z_image_shape: int,
    y_image_shape: int,
    x_image_shape: int,
    c_image_shape: int,
    nr_valid_beads: int,
    nr_edge_beads: int,
    nr_out_of_focus_beads: int,
    nr_clustering_beads: int,
    min_distance_z: int,
    min_distance_y: int,
    min_distance_x: int,
    sigma_z: float,
    sigma_y: float,
    sigma_x: float,
    target_min_intensity: float,
    target_max_intensity: float,
    do_noise: bool,
    signal: int,
    dtype: np.dtype,
):
    # Generate the image as float64
    image = np.zeros(
        shape=(z_image_shape, y_image_shape, x_image_shape, c_image_shape),
        dtype="float32",
    )

    applied_sigmas = []
    non_edge_bead_positions = []
    edge_bead_positions = []
    valid_bead_positions = []
    out_of_focus_bead_positions = []
    clustering_bead_positions = []

    # The strategy is as follows:
    # 1. Generate the valid beads in the center of the image.
    # Those equal to valid_beads + out_of_focus_beads + clustering_beads
    # 2. Generate the edge beads in the edge of the image making sure that they are not too close to the valid beads
    # 3. Gradually remove out_of_focus_beads and clustering_beads from those not in the edge
    while len(non_edge_bead_positions) < (
        nr_valid_beads + nr_out_of_focus_beads + nr_clustering_beads
    ):
        z_pos = z_image_shape // 2
        y_pos = random.randint(min_distance_y + 2, y_image_shape - min_distance_y - 2)
        x_pos = random.randint(min_distance_x + 2, x_image_shape - min_distance_x - 2)
        if not non_edge_bead_positions:
            non_edge_bead_positions.append((z_pos, y_pos, x_pos))
        for pos in non_edge_bead_positions:
            if abs(pos[1] - y_pos) <= min_distance_y and abs(pos[2] - x_pos) <= min_distance_x:
                break
            else:
                continue
        else:
            non_edge_bead_positions.append((z_pos, y_pos, x_pos))

    while len(edge_bead_positions) < nr_edge_beads:
        z_pos = z_image_shape // 2
        y_pos = random.choice(
            [
                random.randint(5, min_distance_y // 2 - 2),
                random.randint(y_image_shape - min_distance_y // 2 + 2, y_image_shape - 5),
            ]
        )
        x_pos = random.choice(
            [
                random.randint(5, min_distance_x // 2 - 2),
                random.randint(x_image_shape - min_distance_x // 2 + 2, x_image_shape - 5),
            ]
        )
        if not edge_bead_positions:
            edge_bead_positions.append((z_pos, y_pos, x_pos))
        for pos in edge_bead_positions:
            if abs(pos[1] - y_pos) <= min_distance_y and abs(pos[2] - x_pos) <= min_distance_x:
                break
            else:
                continue
        else:
            edge_bead_positions.append((z_pos, y_pos, x_pos))

    for _ in range(nr_out_of_focus_beads):
        pos = non_edge_bead_positions.pop()
        pos = (
            random.choice(
                [
                    random.randint(3, min_distance_z - 2),
                    random.randint(z_image_shape - min_distance_z + 2, z_image_shape - 4),
                ]
            ),
            pos[1],
            pos[2],
        )
        out_of_focus_bead_positions.append(pos)

    for _ in range(nr_clustering_beads):
        pos_1 = non_edge_bead_positions.pop()
        pos_2 = (
            pos_1[0],
            pos_1[1] + random.choice([-1, 1]),
            pos_1[2] + random.choice([-1, 1]),
        )
        image[pos_1[0], pos_1[1], pos_1[2], :] = np.random.normal(signal * 1.5, signal / 10)
        image[pos_2[0], pos_2[1], pos_2[2], :] = np.random.normal(signal * 1.5, signal / 10)
        clustering_bead_positions.append(
            (pos_1[0], (pos_1[1] + pos_2[1]) // 2, (pos_1[2] + pos_2[2]) // 2)
        )

    # Fill the image with the beads and adding some normal distributed random intensity
    for pos in edge_bead_positions:
        image[pos[0], pos[1], pos[2], :] = np.random.normal(signal, signal / 50)
    for pos in non_edge_bead_positions:
        image[pos[0], pos[1], pos[2], :] = np.random.normal(signal, signal / 50)
        valid_bead_positions.append(pos)
    for pos in out_of_focus_bead_positions:
        image[pos[0], pos[1], pos[2], :] = np.random.normal(signal, signal / 50)

    # Apply a gaussian filter to the image
    for ch in range(c_image_shape):
        sigma_correction = 1 + ch * 0.1
        applied_sigmas.append(
            (
                sigma_z * sigma_correction,
                sigma_y * sigma_correction,
                sigma_x * sigma_correction,
            )
        )
        image[:, :, :, ch] = skimage_gaussian(
            image[:, :, :, ch], sigma=applied_sigmas[-1], preserve_range=True
        )

    image = skimage_rescale_intensity(
        image,
        in_range=(0.0, signal / 20),
        out_range=(target_min_intensity, target_max_intensity),
    )

    # Add noise
    if do_noise:
        image = skimage_random_noise(image, mode="poisson", clip=False)

    image = skimage_rescale_intensity(image, in_range=(0.0, 1.0), out_range=dtype)

    image = np.expand_dims(image, 0)

    return (
        image,
        applied_sigmas,
        non_edge_bead_positions,
        edge_bead_positions,
        valid_bead_positions,
        out_of_focus_bead_positions,
        clustering_bead_positions,
    )


@st.composite
def st_psf_beads_test_data(
    draw,
    nr_images=st.integers(min_value=1, max_value=3),
    # We want an odd number of slices, so we can have a center slice
    z_image_shape=st.integers(min_value=51, max_value=71).filter(lambda x: x % 2 != 0),
    y_image_shape=st.integers(min_value=512, max_value=1024),
    x_image_shape=st.integers(min_value=512, max_value=1024),
    c_image_shape=st.integers(min_value=1, max_value=3),
    dtype=st.sampled_from([np.uint8, np.uint16]),
    do_noise=st.just(True),
    signal=st.floats(min_value=20.0, max_value=1000.0),
    target_min_intensity=st.floats(min_value=0.01, max_value=0.02),
    target_max_intensity=st.floats(min_value=0.5, max_value=0.9),
    sigma_z=st.floats(min_value=0.9, max_value=1.5),
    sigma_x=st.floats(min_value=0.9, max_value=1.5),
    sigma_y=st.floats(min_value=0.9, max_value=1.5),
    min_distance=st.just(20),
    nr_valid_beads=st.integers(min_value=1, max_value=10),
    nr_edge_beads=st.integers(min_value=0, max_value=3),
    nr_out_of_focus_beads=st.integers(min_value=0, max_value=3),
    nr_clustering_beads=st.integers(min_value=0, max_value=3),
):
    output = {
        "images": [],
        "valid_bead_positions": [],
        "edge_bead_positions": [],
        "out_of_focus_bead_positions": [],
        "clustering_bead_positions": [],
        "applied_sigmas": [],
        "min_distance_z": [],
        "min_distance_y": [],
        "min_distance_x": [],
        "signal": [],
        "do_noise": [],
    }

    z_image_shape = draw(z_image_shape)
    y_image_shape = draw(y_image_shape)
    x_image_shape = draw(x_image_shape)
    c_image_shape = draw(c_image_shape)

    do_noise = draw(do_noise)

    dtype = draw(dtype)

    min_distance_z = draw(min_distance) // 2
    min_distance_y = draw(min_distance)
    min_distance_x = draw(min_distance)

    for _ in range(draw(nr_images)):
        _nr_valid_beads = draw(nr_valid_beads)
        _nr_edge_beads = draw(nr_edge_beads)
        _nr_out_of_focus_beads = draw(nr_out_of_focus_beads)
        _nr_clustering_beads = draw(nr_clustering_beads)
        # We want at least one bead and not too many beads
        assume(
            20
            > (_nr_valid_beads + _nr_edge_beads + _nr_out_of_focus_beads + _nr_clustering_beads)
            > 0
        )

        _signal = draw(signal)
        _target_min_intensity = draw(target_min_intensity)
        _target_max_intensity = draw(target_max_intensity)

        _sigma_z = draw(sigma_z)
        _sigma_y = draw(sigma_y)
        _sigma_x = draw(sigma_x)

        # We do not want images that are too elongated
        assume(0.5 < (x_image_shape / y_image_shape) < 2)

        (
            image,
            applied_sigmas,
            non_edge_bead_positions,
            edge_bead_positions,
            valid_bead_positions,
            out_of_focus_bead_positions,
            clustering_bead_positions,
        ) = _gen_psf_beads_image(
            z_image_shape=z_image_shape,
            y_image_shape=y_image_shape,
            x_image_shape=x_image_shape,
            c_image_shape=c_image_shape,
            nr_valid_beads=_nr_valid_beads,
            nr_edge_beads=_nr_edge_beads,
            nr_out_of_focus_beads=_nr_out_of_focus_beads,
            nr_clustering_beads=_nr_clustering_beads,
            min_distance_z=min_distance_z,
            min_distance_y=min_distance_y,
            min_distance_x=min_distance_x,
            sigma_z=_sigma_z,
            sigma_y=_sigma_y,
            sigma_x=_sigma_x,
            target_min_intensity=_target_min_intensity,
            target_max_intensity=_target_max_intensity,
            do_noise=do_noise,
            signal=_signal,
            dtype=dtype,
        )

        output["images"].append(image)
        output["valid_bead_positions"].append(valid_bead_positions)
        output["edge_bead_positions"].append(edge_bead_positions)
        output["out_of_focus_bead_positions"].append(out_of_focus_bead_positions)
        output["clustering_bead_positions"].append(clustering_bead_positions)
        output["applied_sigmas"].append(applied_sigmas)
        output["min_distance_z"].append(min_distance_z)
        output["min_distance_y"].append(min_distance_y)
        output["min_distance_x"].append(min_distance_x)
        output["signal"].append(_signal)
        output["do_noise"].append(do_noise)

    return output


@st.composite
def st_psf_beads_dataset(
    draw,
    unprocessed_dataset=st_mm_analyses_schema.st_mm_psf_beads_unprocessed_dataset(),
    test_data=st_psf_beads_test_data(),
):
    test_data = draw(test_data)
    psf_beads_unprocessed_dataset = draw(unprocessed_dataset)

    psf_beads_unprocessed_dataset.input_data.psf_beads_images = [
        numpy_to_mm_image(
            array=image,
            name=f"PSF_image_{i}",
            channel_names=[f"Channel_{c}" for c in range(image.shape[-1])],
        )
        for i, image in enumerate(test_data.pop("images"))
    ]

    # Setting the bit depth to the data type of the image
    image_dtype = {
        a.array_data.dtype for a in psf_beads_unprocessed_dataset.input_data.psf_beads_images
    }
    if len(image_dtype) != 1:
        raise ValueError("All images should have the same data type")
    image_dtype = image_dtype.pop()
    if np.issubdtype(image_dtype, np.integer):
        psf_beads_unprocessed_dataset.input_parameters.bit_depth = np.iinfo(image_dtype).bits
    elif np.issubdtype(image_dtype, np.floating):
        psf_beads_unprocessed_dataset.input_parameters.bit_depth = np.finfo(image_dtype).bits
    else:
        psf_beads_unprocessed_dataset.input_parameters.bit_depth = None

    return {
        "unprocessed_dataset": psf_beads_unprocessed_dataset,
        "expected_output": test_data,
    }



"""These are some possibly useful code snippets"""

import logging
from itertools import permutations
from typing import List, Optional

import numpy as np
import pandas as pd
from scipy import ndimage, special
from scipy.optimize import curve_fit, fsolve
from scipy.spatial.distance import cdist
from skimage.feature import peak_local_max
from skimage.filters import apply_hysteresis_threshold, gaussian, threshold_otsu
from skimage.measure import label, regionprops
from skimage.morphology import ball, closing, cube, octahedron, square
from skimage.segmentation import clear_border

# Creating logging services
logger = logging.getLogger(__name__)

INT_DETECTOR_BIT_DEPTHS = [8, 10, 11, 12, 15, 16, 32]
FLOAT_DETECTOR_BIT_DEPTHS = [32, 64]


def airy_fun(
    x: np.ndarray, centre: np.float64, amp: np.float64
) -> np.ndarray:  # , exp):  # , amp, bg):
    with np.errstate(divide="ignore", invalid="ignore"):
        return np.where(
            (x - centre) == 0,
            amp * 0.5**2,
            amp * (special.j1(x - centre) / (x - centre)) ** 2,
        )


def gaussian_fun(x, background, amplitude, center, sd):
    gauss = np.exp(-np.power(x - center, 2.0) / (2 * np.power(sd, 2.0)))
    return background + (amplitude - background) * gauss


def fit_gaussian(profile, guess=None):
    if guess is None:
        guess = [profile.min(), profile.max(), profile.argmax(), 0.8]
    x = np.linspace(0, profile.shape[0], profile.shape[0], endpoint=False)
    popt, pcov, infodict, mesg, ier = curve_fit(
        f=gaussian_fun, xdata=x, ydata=profile, p0=guess, maxfev=5000, full_output=True
    )

    if ier not in [1, 2, 3, 4]:
        logger.warning(f"No gaussian fit found. Reason: {mesg}")

    fitted_profile = gaussian_fun(x, popt[0], popt[1], popt[2], popt[3])
    fwhm = popt[3] * 2.35482

    # Calculate the fit quality using the coefficient of determination (R^2)
    y_mean = np.mean(profile)
    sst = np.sum((profile - y_mean) ** 2)
    ssr = np.sum((profile - fitted_profile) ** 2)
    cd_r2 = 1 - (ssr / sst)

    return fitted_profile, cd_r2, fwhm, popt


def fit_airy(profile, guess=None):
    profile = (profile - profile.min()) / (profile.max() - profile.min())
    if guess is None:
        guess = [profile.argmax(), 4 * profile.max()]
    x = np.linspace(0, profile.shape[0], profile.shape[0], endpoint=False)
    popt, pcov, infodict, mesg, ier = curve_fit(
        f=airy_fun, xdata=x, ydata=profile, p0=guess, full_output=True
    )

    if ier not in [1, 2, 3, 4]:
        logger.warning(f"No airy fit found. Reason: {mesg}")

    fitted_profile = airy_fun(x, popt[0], popt[1])

    # Calculate the FWHM
    def _f(d):
        return airy_fun(d, popt[0], popt[1]) - (fitted_profile.max() - fitted_profile.min()) / 2

    guess = np.array([fitted_profile.argmax() - 1, fitted_profile.argmax() + 1])
    v = fsolve(_f, guess)
    fwhm = abs(v[1] - v[0])

    # Calculate the fit quality using the coefficient of determination (R^2)
    y_mean = np.mean(profile)
    sst = np.sum((profile - y_mean) ** 2)
    ssr = np.sum((profile - fitted_profile) ** 2)
    cd_r2 = 1 - (ssr / sst)

    return fitted_profile, cd_r2, fwhm, popt


def multi_airy_fun(x: np.ndarray, *params) -> np.ndarray:
    y = np.zeros_like(x)
    for i in range(0, len(params), 2):
        y = y + airy_fun(x, params[i], params[i + 1])
    return y


def is_saturated(
    channel: np.ndarray,
    threshold: float = 0.0,
    detector_bit_depth: Optional[int] = None,
) -> bool:
    """
    Checks if the channel is saturated.
    A warning if it suspects that the detector bit depth does not match the datatype.
    thresh: float
        Threshold for the ratio of saturated pixels to total pixels
    detector_bit_depth: int
        Bit depth of the detector. Sometimes, detectors bit depth are not matching the datatype of the measureemnts.
        Here it can be specified the bit depth of the detector if known. The function is going to raise
        If None, it will be inferred from the channel dtype.
    """
    if detector_bit_depth is not None:
        if (
            detector_bit_depth not in INT_DETECTOR_BIT_DEPTHS
            and detector_bit_depth not in FLOAT_DETECTOR_BIT_DEPTHS
        ):
            raise ValueError(
                f"The detector bit depth provided ({detector_bit_depth}) is not supported. Supported values are {INT_DETECTOR_BIT_DEPTHS} for integer detectors and {FLOAT_DETECTOR_BIT_DEPTHS} for floating point detectors."
            )
        if (
            np.issubdtype(channel.dtype, np.integer)
            and detector_bit_depth not in INT_DETECTOR_BIT_DEPTHS
        ):
            raise ValueError(
                f"The channel datatype {channel.dtype} does not match the detector bit depth {detector_bit_depth}. The channel might be saturated."
            )
        elif (
            np.issubdtype(channel.dtype, np.floating)
            and detector_bit_depth not in FLOAT_DETECTOR_BIT_DEPTHS
        ):
            raise ValueError(
                f"The channel datatype {channel.dtype} does not match the detector bit depth {detector_bit_depth}. The channel might be saturated."
            )
        else:
            if np.issubdtype(channel.dtype, np.integer):
                if detector_bit_depth > np.iinfo(channel.dtype).bits:
                    raise ValueError(
                        f"The channel datatype {channel.dtype} does not support the detector bit depth {detector_bit_depth}. The channel might be saturated."
                    )
                else:
                    max_limit = pow(2, detector_bit_depth) - 1
            elif np.issubdtype(channel.dtype, np.floating):
                if detector_bit_depth != np.finfo(channel.dtype).bits:
                    raise ValueError(
                        f"The channel datatype {channel.dtype} does not support the detector bit depth {detector_bit_depth}. The channel might be saturated."
                    )
                else:
                    max_limit = np.finfo(channel.dtype).max

    else:
        if np.issubdtype(channel.dtype, np.integer):
            max_limit = np.iinfo(channel.dtype).max
        elif np.issubdtype(channel.dtype, np.floating):
            max_limit = np.finfo(channel.dtype).max
        else:
            raise ValueError("The channel provided is not a valid numpy dtype.")

    if channel.max() > max_limit:
        raise ValueError(
            "The channel provided has values larger than the bit depth of the detector."
        )

    saturation_matrix = channel == max_limit
    saturation_ratio = np.count_nonzero(saturation_matrix) / channel.size

    return saturation_ratio > threshold


def _segment_channel(
    channel,
    min_distance,
    method,
    threshold,
    sigma,
    low_corr_factor,
    high_corr_factor,
):
    """Segment a channel (3D numpy array ZYX)"""
    if threshold is None:
        threshold = threshold_otsu(channel)

    if sigma is not None:
        channel = gaussian(image=channel, sigma=sigma, preserve_range=True, channel_axis=None)

    if method == "hysteresis":  # We may try here hysteresis threshold
        thresholded = apply_hysteresis_threshold(
            channel, low=threshold * low_corr_factor, high=threshold * high_corr_factor
        )

    elif method == "local_max":  # We are applying a local maxima algorithm
        peaks = peak_local_max(
            channel,
            min_distance=min_distance,
            threshold_abs=(threshold * 0.5),
        )
        thresholded = np.copy(channel)
        thresholded[tuple(peaks.T)] = thresholded.max()
        thresholded = apply_hysteresis_threshold(
            thresholded,
            low=threshold * low_corr_factor,
            high=threshold * high_corr_factor,
        )
    else:
        raise Exception("A valid segmentation method was not provided")

    closed = closing(thresholded, cube(min_distance))
    cleared = clear_border(closed)
    return label(cleared)


def segment_image(
    image: np.ndarray,
    min_distance: float,
    threshold: float = None,
    sigma=None,
    method: str = "local_max",
    low_corr_factors: List[float] = None,
    high_corr_factors: List[float] = None,
):
    """Segment an image and return a labels object.
    Image must be provided as TZYXC numpy array
    """
    logger.info("Image being segmented...")

    if low_corr_factors is None or (
        isinstance(low_corr_factors, list) and len(low_corr_factors) == 0
    ):
        low_corr_factors = [0.95] * image.shape[4]
        logger.warning("No low correction factor specified. Using defaults")
    if high_corr_factors is None or (
        isinstance(high_corr_factors, list) and len(high_corr_factors) == 0
    ):
        high_corr_factors = [1.05] * image.shape[4]
        logger.warning("No high correction factor specified. Using defaults")

    if len(high_corr_factors) != image.shape[4] or len(low_corr_factors) != image.shape[4]:
        raise Exception("The number of correction factors does not match the number of channels.")

    # We create an empty array to store the output
    labels_image = np.zeros(image.shape, dtype=np.uint16)
    for c in range(image.shape[4]):
        for t in range(image.shape[0]):
            labels_image[t, :, :, :, c] = _segment_channel(
                image[t, :, :, :, c],
                min_distance=min_distance,
                method=method,
                threshold=threshold,
                sigma=sigma,
                low_corr_factor=low_corr_factors[c],
                high_corr_factor=high_corr_factors[c],
            )
    return labels_image


def _compute_channel_spots_properties(
    channel, label_channel, remove_center_cross=False, pixel_size=None
):
    """Analyzes and extracts the properties of a single channel"""

    regions = regionprops(label_channel, channel)

    ch_properties = [
        {
            "label": region.label,
            "area": region.area,
            "centroid": region.centroid,
            "weighted_centroid": region.weighted_centroid,
            "max_intensity": region.max_intensity,
            "mean_intensity": region.mean_intensity,
            "min_intensity": region.min_intensity,
            "integrated_intensity": region.mean_intensity * region.area,
        }
        for region in regions
    ]

    if (
        remove_center_cross
    ):  # Argolight spots pattern contains a central cross that we might want to remove
        largest_area = 0
        largest_region = None
        for region in ch_properties:
            if region["area"] > largest_area:  # We assume the cross is the largest area
                largest_area = region["area"]
                largest_region = region
        if largest_region:
            ch_properties.remove(largest_region)
    ch_positions = np.array([x["weighted_centroid"] for x in ch_properties])
    if pixel_size:
        ch_positions = ch_positions[0:] * pixel_size

    return ch_properties, ch_positions


def compute_spots_properties(image, labels, remove_center_cross=False, pixel_size=None):
    """Computes a number of properties for the PSF-like spots found on an image provided they are segmented.
    Image must be provided as a TZYXC 5d numpy array"""
    properties = []
    positions = []

    for c in range(image.shape[4]):
        for t in range(image.shape[0]):
            pr, pos = _compute_channel_spots_properties(
                channel=image[t, :, :, :, c],
                label_channel=labels[t, :, :, :, c],
                remove_center_cross=remove_center_cross,
                pixel_size=pixel_size,
            )
            properties.append(pr)
            positions.append(pos)

    return properties, positions


def compute_distances_matrix(positions, max_distance, pixel_size=None):
    """Calculates Mutual Closest Neighbour distances between all channels and returns the values as"""
    logger.info("Computing distances between spots")

    if len(positions) < 2:
        raise Exception("Not enough dimensions to do a distance measurement")

    channel_permutations = list(permutations(range(len(positions)), 2))

    if not pixel_size:  # TODO: make sure the units are corrected if no pixel size
        pixel_size = np.array((1, 1, 1))
        logger.warning("No pixel size specified. Using the unit")
    else:
        pixel_size = np.array(pixel_size)

    distances_rows = []

    for a, b in channel_permutations:
        distances_matrix = cdist(positions[a], positions[b], w=pixel_size)

        distances_rows.extend(
            {
                "channel_a": a,
                "channel_b": b,
                "z_coord_a": pos_A[0],
                "y_coord_a": pos_A[1],
                "x_coord_a": pos_A[2],
                "z_coord_b": positions[b][d.argmin()][0],
                "y_coord_b": positions[b][d.argmin()][1],
                "x_coord_b": positions[b][d.argmin()][2],
                "z_dist": pos_A[0] - positions[b][d.argmin()][0],
                "y_dist": pos_A[1] - positions[b][d.argmin()][1],
                "x_dist": pos_A[2] - positions[b][d.argmin()][2],
                "dist_3d": d.min(),
                "labels_a": i,
                "labels_b": d.argmin(),
            }
            for i, (pos_A, d) in enumerate(zip(positions[a], distances_matrix))
            if d.min() < max_distance
        )

    return pd.DataFrame(distances_rows)


def _radial_mean(image, bins=None):
    """Computes the radial mean from an input 2d image.
    Taken from scipy-lecture-notes 2.6.8.4
    """
    # TODO: workout a binning = image size
    if not bins:
        bins = 200
    size_x, size_y = image.shape
    x, y = np.ogrid[0:size_x, 0:size_y]

    r = np.hypot(x - size_x / 2, y - size_y / 2)

    rbin = (bins * r / r.max()).astype(np.int)

    return ndimage.mean(image, labels=rbin, index=np.arange(1, rbin.max() + 1))


def _channel_fft_2d(channel):
    channel_fft = np.fft.rfft2(channel)
    return np.fft.fftshift(np.abs(channel_fft), axes=1)


def fft_2d(image):
    # Create an empty array to contain the transform
    fft = np.zeros(shape=(image.shape[1], image.shape[2], image.shape[3] // 2 + 1), dtype="float64")
    for c in range(image.shape[2]):
        fft[c, :, :] = _channel_fft_2d(image[..., c, :, :])

    return fft


def _channel_fft_3d(channel):
    channel_fft = np.fft.rfftn(channel)
    return np.fft.fftshift(np.abs(channel_fft), axes=1)


def fft_3d(image):
    fft = np.zeros(
        shape=(
            image.shape[0],
            image.shape[1],
            image.shape[2],
            image.shape[3],
            image.shape[4] // 2 + 1,
        ),  # We only compute the real part of the FFT
        dtype="float64",
    )
    for c in range(image.shape[-3]):
        fft[..., c, :, :] = _channel_fft_3d(image[..., c, :, :])

    return fft


def wavelength_to_rgb(wavelength, gamma=0.8):
    """
    Copied from https://www.noah.org/wiki/Wavelength_to_RGB_in_Python
    This converts a given wavelength of light to an
    approximate RGB color value. The wavelength must be given
    in nanometers in the range from 380 nm through 750 nm
    (789 THz through 400 THz).

    Based on code by Dan Bruton
    http://www.physics.sfasu.edu/astro/color/spectra.html
    """

    wavelength = float(wavelength)
    if 380 < wavelength <= 440:
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        r = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        g = 0.0
        b = (1.0 * attenuation) ** gamma
    elif 440 < wavelength <= 490:
        r = 0.0
        g = ((wavelength - 440) / (490 - 440)) ** gamma
        b = 1.0
    elif 490 < wavelength <= 510:
        r = 0.0
        g = 1.0
        b = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif 510 < wavelength <= 580:
        r = ((wavelength - 510) / (580 - 510)) ** gamma
        g = 1.0
        b = 0.0
    elif 580 < wavelength <= 645:
        r = 1.0
        g = (-(wavelength - 645) / (645 - 580)) ** gamma
        b = 0.0
    elif 645 < wavelength <= 750:
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        r = (1.0 * attenuation) ** gamma
        g = 0.0
        b = 0.0
    else:
        r = 0.0
        g = 0.0
        b = 0.0
    r *= int(r * 255)
    g *= int(g * 255)
    g *= int(g * 255)
    return r, g, b
